static STT_BUILTIN(sb_2num)
{
	stt_obj_t *obj;
	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "->num needs a parameter");

	switch(obj->type) {
		case STT_ST_INT:
			return stt_create_num(ctx->stt, obj->d.integer.value);
		case STT_ST_NUM:
		case STT_ST_ERROR:
			return obj;
		case STT_ST_STRING:
			return stt_create_num(ctx->stt, atof(obj->d.string.value));
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Type can't be converted to number");
	}
}

static STT_BUILTIN(sb_2int)
{
	stt_obj_t *obj;
	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "->int needs a parameter");

	switch(obj->type) {
		case STT_ST_NUM:
			return stt_create_int(ctx->stt, (int) obj->d.num.value);
		case STT_ST_INT:
		case STT_ST_ERROR:
			return obj;
		case STT_ST_STRING:
			return stt_create_int(ctx->stt, atoi(obj->d.string.value));
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Type can't be converted to integer");
	}
}

static STT_BUILTIN(sb_2str)
{
	stt_obj_t *obj;
	char buf[32], *p;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "->str needs a parameter");

	switch(obj->type) {
		case STT_ST_STRING:
		case STT_ST_ERROR:
			return obj;
		case STT_ST_SYMBOL:
			if(obj->d.symbol.colon) {
				p = (char *) stt_malloc(strlen(obj->d.symbol.name) + 1);
				strcpy(p, ":");
				strcat(p, obj->d.symbol.name);
				return stt_create_string_allocd(ctx->stt, p);
			} else
				return stt_create_string(ctx->stt, obj->d.symbol.name);
		case STT_ST_INT:
			sprintf(buf, "%ld", obj->d.integer.value);
			return stt_create_string(ctx->stt, buf);
		case STT_ST_CHAR:
			buf[0] = obj->d.ch.ch;
			buf[1] = 0;
			return stt_create_string(ctx->stt, buf);
		case STT_ST_NUM:
			sprintf(buf, "%f", obj->d.num.value);
			return stt_create_string(ctx->stt, buf);
		case STT_ST_NIL:
			return stt_create_string(ctx->stt, "nil");
		case STT_ST_BUILTIN:
			if(obj->d.builtin.symbol)
				return stt_create_string(ctx->stt, obj->d.builtin.symbol);
			else
				return stt_create_string(ctx->stt, "#anonymous-builtin");
			break;
		case STT_ST_T:
			return stt_create_string(ctx->stt, "T");
		default:
			return stt_create_error_with_obj(ctx->stt, STT_SE_TYPE, "Type can't be converted to string", obj);
	}
}

static STT_BUILTIN(sb_2symbol)
{
	stt_obj_t *obj;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "->symbol needs a parameter");

	switch(obj->type) {
		case STT_ST_SYMBOL:
		case STT_ST_ERROR:
			return obj;
		case STT_ST_STRING:
			if(obj->d.string.value[0] == ':') {
				obj = stt_create_symbol(ctx->stt, obj->d.string.value + 1);
				obj->d.symbol.colon = 1;
				return obj;
			} else
				return stt_create_symbol(ctx->stt, obj->d.string.value);
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Type can't be converted to symbol");
	}
}

static STT_BUILTIN(sb_random)
{
	stt_obj_t *obj, *o;
	int i;
	double r;

	if(data == 0) {
		/* random */
		if(!stt_nextarg(ctx, &obj, &parm))
			return stt_create_error(ctx->stt, STT_SE_PNUM, "random needs a parameter");
		switch(obj->type) {
			case STT_ST_ERROR: return obj;
			case STT_ST_INT:
				if(obj->d.integer.value == 0)
					return stt_create_error(ctx->stt, STT_SE_PARM, "Integer parameter for random can't be zero");
				if(obj->d.integer.value < 0)
					return stt_create_int(ctx->stt, - rand() % obj->d.integer.value);
				else
					return stt_create_int(ctx->stt, rand() % obj->d.integer.value);
			case STT_ST_CONS:
				/* be clever: return random element of list */
				for(o=obj,i=0; o->type == STT_ST_CONS; o=o->d.cons.cdr, i++);
				i = rand() % i;
				for(o=obj; i>0; o=o->d.cons.cdr, i--);
				return o->d.cons.car;
			default: return stt_create_error(ctx->stt, STT_SE_TYPE, "Parameter for random must be an integer (or a CONS)");
		}
	} else {
		r = (double) rand() / RAND_MAX;
		if(stt_nextarg(ctx, &obj, &parm)) {
			if(obj->type == STT_ST_ERROR) return obj;
			if(obj->type != STT_ST_NUM)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "Parameter for rand must be a float");
			return stt_create_num(ctx->stt, r * obj->d.num.value);
		}
		return stt_create_num(ctx->stt, r);
	}
}

/* math helper: performs primitive algebraic functions with number polymorphism */
static stt_obj_t *alg_perform(stt_varctx_t *ctx, stt_obj_t *obj, stt_obj_t *arg, int func)
{
	int dtype;
	long obj_int;
	double obj_num;
	long arg_int;
	double arg_num;

	dtype = obj->type;
	switch(dtype) {
		case STT_ST_INT:
			switch(arg->type) {
				case STT_ST_NUM:
					obj_num = obj->d.integer.value;
					arg_num = arg->d.num.value;
					dtype = STT_ST_NUM;
					break;
				case STT_ST_INT:
					obj_int = obj->d.integer.value;
					arg_int = arg->d.integer.value;
					break;
				default:
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Type mismatch (algebra arg 2 after integer)");

			}
			break;
		case STT_ST_NUM:
			obj_num = obj->d.num.value;
			switch(arg->type) {
				case STT_ST_INT:
					arg_num = arg->d.integer.value;
					break;
				case STT_ST_NUM:
					arg_num = arg->d.num.value;
					break;
				default:
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Type mismatch (algebra arg 2 after float)");
			}
			break;
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Type mismatch (algebra arg 1)");
	}

	switch(dtype) {
		case STT_ST_INT: /* perform on integers */
			switch(func) {
				case '+':
					return stt_create_int(ctx->stt, obj_int + arg_int);
				case '-':
					return stt_create_int(ctx->stt, obj_int - arg_int);
				case '*':
					return stt_create_int(ctx->stt, obj_int * arg_int);
				case '/':
					if(arg_int == 0) return stt_create_error(ctx->stt, STT_SE_MATH, "Division by zero");
					return stt_create_int(ctx->stt, obj_int / arg_int);
				case '%':
					if(arg_int == 0) return stt_create_error(ctx->stt, STT_SE_MATH, "Division by zero");
					return stt_create_int(ctx->stt, obj_int % arg_int);

				case '<': /* predicate support: return rvalue on true, NULL on false */
					if(obj_int < arg_int)
						return arg;
					else
						return NULL;

				case '>':
					if(obj_int > arg_int)
						return arg;
					else
						return NULL;

				case '{':
					if(obj_int <= arg_int)
						return arg;
					else
						return NULL;

				case '}':
					if(obj_int >= arg_int)
						return arg;
					else
						return NULL;

				case '=':
					if(obj_int == arg_int)
						return arg;
					else
						return NULL;

				case '&': /* ..let's support bitwise logical operations while we're at it */
					return stt_create_int(ctx->stt, obj_int & arg_int);
				case '|':
					return stt_create_int(ctx->stt, obj_int | arg_int);
				case '^':
					return stt_create_int(ctx->stt, obj_int ^ arg_int);
				default:
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Function not supported on integers");
			}

			break;
		case STT_ST_NUM: /* perform on floats */
			switch(func) {
				case '+':
					return stt_create_num(ctx->stt, obj_num + arg_num);
				case '-':
					return stt_create_num(ctx->stt, obj_num - arg_num);
				case '*':
					return stt_create_num(ctx->stt, obj_num * arg_num);
				case '/':
/*					if(arg_num == 0) return stt_create_error(ctx->stt, STT_SE_MATH, "Division by zero"); */
					return stt_create_num(ctx->stt, obj_num / arg_num);
				case '%':
/*					if(arg_num == 0) return stt_create_error(ctx->stt, STT_SE_MATH, "Division by zero"); */
					return stt_create_num(ctx->stt, fmod(obj_num, arg_num));

				case '<': /* predicate support: return rvalue on true, NULL on false */
					if(obj_num < arg_num) {
						if(arg->type == STT_ST_INT)
							return stt_create_num(ctx->stt, arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '>':
					if(obj_num > arg_num) {
						if(arg->type == STT_ST_INT)
							return stt_create_num(ctx->stt, arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '{':
					if(obj_num <= arg_num) {
						if(arg->type == STT_ST_INT)
							return stt_create_num(ctx->stt, arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '}':
					if(obj_num >= arg_num) {
						if(arg->type == STT_ST_INT)
							return stt_create_num(ctx->stt, arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				case '=':
					if(obj_num == arg_num) {
						if(arg->type == STT_ST_INT)
							return stt_create_num(ctx->stt, arg_num);
						else
							return arg;
					} else {
						return NULL;
					}

				default:
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Function not supported on floats");
			}

		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Algebra: Type mismatch");
	}
}

static STT_BUILTIN(sb_lognot)
{
	stt_obj_t *obj;
	STT_ARG_INT(ctx->stt, obj);

	return stt_create_int(ctx->stt, ~obj->d.integer.value);
}

static STT_BUILTIN(sb_logic)
{
	stt_obj_t *value, *arg;
	int op;
	switch((stt_intptr_t) data) {
		case 0:
		default:
			value = stt_create_int(ctx->stt, 0);
			op = '|';
			break;
		case 1:
			value = stt_create_int(ctx->stt, -1);
			op = '&';
			break;
		case 2:
			value = stt_create_int(ctx->stt, 0);
			op = '^';
			break;
	}

	while(stt_nextarg(ctx, &arg, &parm)) {
		if(arg->type == STT_ST_ERROR) return arg;
		value = alg_perform(ctx, value, arg, op);
		if(value->type == STT_ST_ERROR) return value;
	}

	return value;
}

static STT_BUILTIN(sb_plus)
{
	stt_obj_t *value = stt_create_int(ctx->stt, 0);
	stt_obj_t *obj;
	char *str = NULL;
	int strl = 0;
	int stra = 0;
	int le;
	int type = -1;

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) {
			if(str) free(str);
			return obj;
		}

		if(type == -1) {
			type = obj->type;
			if(type == STT_ST_CHAR) {
				strl = 1;
				stra = 3;
				str = (char *) stt_malloc(stra);
				str[0] = obj->d.ch.ch;
				str[1] = 0;
				type = STT_ST_STRING;
				continue;
			}

			if(type != STT_ST_STRING && type != STT_ST_NUM && type != STT_ST_INT)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "+: Type mismatch");
		}

		if(obj->type == STT_ST_STRING || obj->type == STT_ST_CHAR) {
			if(type != STT_ST_STRING) return stt_create_error(ctx->stt, STT_SE_TYPE, "+: Type mismatch, adding to string");

			if(obj->type == STT_ST_CHAR) le = 1; else le = strlen(obj->d.string.value);
			if(!str) {
				strl = le; stra = le * 2 + 1;
				str = (char *) stt_malloc(stra);
				strcpy(str, obj->d.string.value);
			} else {
				if(stra < strl + le + 1) {
					do stra *= 2; while(stra < strl + le + 1);
					str = (char *) stt_realloc(str, stra);
				}

				if(obj->type == STT_ST_STRING) {
					strcat(str, obj->d.string.value);
				} else {
					str[strl] = obj->d.ch.ch;
					str[strl+1] = 0;
				}

				strl += le;
			}

			continue;
		}

		if(type == STT_ST_STRING) return stt_create_error(ctx->stt, STT_SE_TYPE, "+: Type mismatch, adding string to number");

		value = alg_perform(ctx, value, obj, '+');
		if(value->type == STT_ST_ERROR) return value;
	}

	if(type == STT_ST_STRING) {
		if(!str) return stt_create_string(ctx->stt, ""); else {
			obj = stt_create_string_allocd(ctx->stt, str);
			return obj;
		}
	} else return value;
}

static STT_BUILTIN(sb_minus)
{
	stt_obj_t *value;
	stt_obj_t *obj;
	int cnt = 0;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Subtraction requires at least one parameter");
	if(obj->type == STT_ST_ERROR) return obj;

	switch(obj->type) {
		case STT_ST_NUM:
		case STT_ST_INT:
			value = obj;
			break;
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "-: Type mismatch");
	}

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;

		value = alg_perform(ctx, value, obj, '-');
		if(value->type == STT_ST_ERROR) return value;

		cnt = 1;
	}

	if(!cnt) value = alg_perform(ctx, stt_create_int(ctx->stt, 0), value, '-');
	return value;
}

static STT_BUILTIN(sb_mul)
{
	stt_obj_t *value = stt_create_int(ctx->stt, 1);
	long ival;
	int j, l;
	stt_obj_t *obj, *nobj;
	char *str, *c, *d;

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;

		if(obj->type == STT_ST_STRING) {
				/* special case: string by integer multiplication? :) */
				if(!stt_nextarg(ctx, &nobj, &parm))
					return stt_create_error(ctx->stt, STT_SE_TYPE, "*: Type mismatch (string by int)");
				if(nobj->type == STT_ST_ERROR) return nobj;
				if(nobj->type != STT_ST_INT)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "*: Type mismatch (string by int)");

				l = strlen(obj->d.string.value);
				str = (char *) malloc(l * abs(nobj->d.integer.value) + 1);
				if(!str)
					return stt_create_error(ctx->stt, STT_SE_MEM, "Not enough memory for new string?");

				c = str;
				if(nobj->d.integer.value < 0) {
					/* reverse string */

					for(ival=nobj->d.integer.value;ival<0;ival++) {
						d = obj->d.string.value + l - 1;
						for(j=0;j<l;j++)
							*c++ = *d--;
					}
					*c++ = 0;

				} else {
					for(ival=0;ival<nobj->d.integer.value;ival++) {
						d = obj->d.string.value;
						for(j=0;j<l;j++)
							*c++ = *d++;
					}
					*c++ = 0;
				}

				obj = stt_create_string_allocd(ctx->stt, str);
				return obj;
		}

		if(obj->type != STT_ST_INT && obj->type != STT_ST_NUM) return stt_create_error(ctx->stt, STT_SE_TYPE, "*: Type mismatch (arg not number)");

		value = alg_perform(ctx, value, obj, '*');
		if(value->type == STT_ST_ERROR) return value;
	}

	return value;
}

static STT_BUILTIN(sb_div)
{
	stt_obj_t *value;
	stt_obj_t *obj;
	int cnt = 0;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Division requires at least one parameter");
	if(obj->type == STT_ST_ERROR) return obj;

	switch(obj->type) {
		case STT_ST_NUM:
		case STT_ST_INT:
			value = obj;
			break;
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "/: Type mismatch (arg not number)");
	}

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;

		value = alg_perform(ctx, value, obj, '/');
		if(value->type == STT_ST_ERROR) return value;

		cnt = 1;
	}

	if(!cnt) value = alg_perform(ctx, stt_create_num(ctx->stt, 1), value, '/');
	return value;
}

static STT_BUILTIN(sb_mod)
{
	stt_obj_t *value;
	stt_obj_t *obj;
	int cnt = 0;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Modulo requires at least one parameter");
	if(obj->type == STT_ST_ERROR) return obj;

	switch(obj->type) {
		case STT_ST_NUM:
		case STT_ST_INT:
			value = obj;
			break;
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "mod: Type mismatch (arg not number)");
	}

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;

		value = alg_perform(ctx, value, obj, '%');
		if(value->type == STT_ST_ERROR) return value;

		cnt = 1;
	}

	if(!cnt) value = alg_perform(ctx, stt_create_int(ctx->stt, 1), value, '%');
	return value;
}

static STT_BUILTIN(sb_atan2)
{
	stt_obj_t *y, *x;
	STT_ARG_NUM(ctx->stt, y);
	STT_ARG_NUM(ctx->stt, x);
	return stt_create_num(ctx->stt, atan2(y->d.num.value, x->d.num.value));
}

static STT_BUILTIN(sb_pow)
{
	stt_obj_t *y, *x;
	STT_ARG_NUM(ctx->stt, x);
	STT_ARG_NUM(ctx->stt, y);
	return stt_create_num(ctx->stt, pow(x->d.num.value, y->d.num.value));
}

static STT_BUILTIN(sb_frexp)
{
	stt_obj_t *x;
	int exp;
	double r;
	STT_ARG_NUM(ctx->stt, x);
	r = frexp(x->d.num.value, &exp);
	return stt_create_cons(ctx->stt, stt_create_num(ctx->stt, r), stt_create_int(ctx->stt, exp));
}

static STT_BUILTIN(sb_equal)
{
	stt_obj_t *value = NULL, *obj;
	char *str = NULL;
	int res = 0;
	int type;

	STT_ARG_OBJ(ctx->stt, obj);

	switch(obj->type) {
		case STT_ST_INT:
			type = obj->type;
			value = obj;
			break;
		case STT_ST_NUM:
			type = obj->type;
			value = obj;
			break;
		case STT_ST_STRING:
			type = STT_ST_STRING;
			str = obj->d.string.value;
			break;
		case STT_ST_SYMBOL:
			type = STT_ST_SYMBOL;
			str = obj->d.symbol.name;
			break;
		case STT_ST_CHAR:
			type = STT_ST_CHAR;
			value = obj;
			break;
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "=: Type mismatch (unknown type)");
	}

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;

		switch(obj->type) {
			case STT_ST_NUM:
			case STT_ST_INT:
				if(type != STT_ST_NUM && type != STT_ST_INT)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "=: Type mismatch (= num notnum)");

				value = alg_perform(ctx, value, obj, (stt_intptr_t)data);
				if(value && value->type == STT_ST_ERROR) return value;
				res = (value != NULL);
				break;
			case STT_ST_CHAR:
				if(type != STT_ST_CHAR)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "=: Type mismatch (= char notchar)");
				switch((stt_intptr_t) data) {
					case '=': res = (value->d.ch.ch == obj->d.ch.ch); break;
					case '<': res = (value->d.ch.ch < obj->d.ch.ch); break;
					case '>': res = (value->d.ch.ch > obj->d.ch.ch); break;
					case '{': res = (value->d.ch.ch <= obj->d.ch.ch); break;
					case '}': res = (value->d.ch.ch >= obj->d.ch.ch); break;
				}
				break;
			case STT_ST_STRING:
				if(type != STT_ST_STRING)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "=: Type mismatch (= str notstr)");

				switch((stt_intptr_t)data) {
					case '=': res = !strcmp(obj->d.string.value, str); break;
					case '<': res = (strcmp(obj->d.string.value, str) > 0); break;
					case '>': res = (strcmp(obj->d.string.value, str) < 0); break;
					case '{': res = (strcmp(obj->d.string.value, str) >= 0); break;
					case '}': res = (strcmp(obj->d.string.value, str) <= 0); break;
				}
				str = obj->d.string.value;
				break;
			case STT_ST_SYMBOL:
				if(type != STT_ST_SYMBOL)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "=: Type mismatch (= sym notsym)");

				switch((stt_intptr_t)data) {
					case '=': res = !strcmp(obj->d.symbol.name, str); break;
					case '<': res = (strcmp(obj->d.symbol.name, str) > 0); break;
					case '>': res = (strcmp(obj->d.symbol.name, str) < 0); break;
					case '{': res = (strcmp(obj->d.symbol.name, str) >= 0); break;
					case '}': res = (strcmp(obj->d.symbol.name, str) <= 0); break;
				}
				str = obj->d.symbol.name;
				break;
			default:
				;
		}

		if(!res)
			return stt_create_nil(ctx->stt);
	}

	return stt_create_t(ctx->stt);
}

static STT_BUILTIN(sb_isset)
{
	stt_obj_t *name;
	if(!stt_nextarg(ctx, &name, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for isset");
	if(name->type != STT_ST_SYMBOL)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "isset: Type mismatch (not symbol)");

	if(data == 0)
		return stt_varctx_get_s(ctx, name->d.symbol.sdbm, name->d.symbol.id) ? stt_create_t(ctx->stt) : stt_create_nil(ctx->stt);
	else
		return stt_varctx_retr_s(ctx, name->d.symbol.sdbm, name->d.symbol.id) ? stt_create_t(ctx->stt) : stt_create_nil(ctx->stt);
}
