static STT_BUILTIN(sb_set)
{
	stt_obj_t *name, *value = NULL;

	while(parm->type == STT_ST_CONS) {
		stt_nextarg(ctx, &name, &parm);
		if(name->type == STT_ST_ERROR) return name;
		if(name->type != STT_ST_SYMBOL) return stt_create_error(ctx->stt, STT_SE_PARM, "Cannot set non-symbol");

		if(!stt_nextarg(ctx, &value, &parm))
			return stt_create_error(ctx->stt, STT_SE_PARM, "Insufficient parameters for set");
		if(value->type == STT_ST_ERROR) return value;

		if(value->type == STT_ST_FUNC && !value->d.func.symbol)
			value->d.func.symbol = stt_get_symbolname(ctx->stt, name->d.symbol.sdbm, name->d.symbol.id);

		stt_varctx_set_s(ctx, name->d.symbol.sdbm, name->d.symbol.id, value);
	}

	return value ? value : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_setg)
{
	stt_obj_t *name, *value = NULL;
	stt_var_t *var;

	while(parm->type == STT_ST_CONS) {
		stt_nextarg(ctx, &name, &parm);
		if(name->type == STT_ST_ERROR) return name;
		if(name->type != STT_ST_SYMBOL) return stt_create_error(ctx->stt, STT_SE_PARM, "Cannot set non-symbol");

		if(!stt_nextarg(ctx, &value, &parm))
			return stt_create_error(ctx->stt, STT_SE_PARM, "Invalid parameters for setg");
		if(value->type == STT_ST_ERROR) return value;

		while(ctx->parent) {
			var = stt_varctx_retr_s(ctx, name->d.symbol.sdbm, name->d.symbol.id);
			if(var) {
				var->value = value;
				return value;
			}
			ctx = ctx->parent;
		}

		stt_varctx_set_s(ctx, name->d.symbol.sdbm, name->d.symbol.id, value);
	}

	return value ? value : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_defuse)
{
	stt_obj_t *obj;
	stt_obj_t *b;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Defuse requires a parameter");
	if(obj->type != STT_ST_BUILTIN) return obj;

	b = stt_create_builtin(ctx->stt, obj->d.builtin.handler);
	b->d.builtin.data = obj->d.builtin.data;
	b->d.builtin.nosubst = 1;
	return b;
}

static STT_BUILTIN(sb_getbind)
{
	stt_var_t *var;
	stt_obj_t *sym;

	STT_ARG_SYMBOL(ctx->stt, sym);

	var = stt_varctx_get_s(ctx, sym->d.symbol.sdbm, sym->d.symbol.id);
	if(!var) return stt_create_error_with_obj(ctx->stt, STT_SE_UNDEF, "Symbol not bound", sym);
	return var->value;
}

static STT_BUILTIN(sb_regerror)
{
	stt_obj_t *name;
	STT_ARG_STRING(ctx->stt, name);
	if(strlen(name->d.string.value) > 15)
		return stt_create_error(ctx->stt, STT_SE_PARM, "register-error: name of error too long");
	return stt_create_int(ctx->stt, stt_regerrorcode(ctx->stt, name->d.string.value));
}

static STT_BUILTIN(sb_function_code)
{
	stt_obj_t *fun;
	STT_ARG_OBJ(ctx->stt, fun);
	if(fun->type != STT_ST_FUNC)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "function-code: argument not a user function");
	return fun->d.func.exp;
}

static STT_BUILTIN(sb_weakptr_clean)
{
	stt_obj_t *obj;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Weakptr-clean requires a parameter");
	if(obj->type == STT_ST_ERROR) return obj;

	return stt_create_int(ctx->stt, stt_weak_check(obj));
}

static STT_BUILTIN(sb_begin)
{
	stt_obj_t *obj = NULL;
	while(stt_nextargtail(ctx, &obj, &parm, tail)) {
		if(obj->type == STT_ST_ERROR) return obj;
	}

	return obj ? obj : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_funcall)
{
	stt_obj_t *func, *parml;

	/* get function */
	STT_ARG_OBJ(ctx->stt, func);

	switch(func->type) {
	case STT_ST_FUNC:
	case STT_ST_GENERIC:
		if((stt_intptr_t) data == 2) {
			/* expand */
			STT_ARG_OBJ(ctx->stt, parml);
			return stt_funcexec(ctx, parml, func, parml, tail, 1);
		} else {
			return stt_funcexec(ctx, parm, func, parm, tail, 0);
		}
	case STT_ST_BUILTIN:
		return func->d.builtin.handler(ctx, parm, func->d.builtin.data, tail);
	default:
		return stt_create_error(ctx->stt, STT_SE_WFUNC, "Invalid function given to funcall.");
	}
}

static STT_BUILTIN(sb_apply)
{
	stt_varctx_t *tctx;
	stt_obj_t *i, *func;

	/* get function */
	STT_ARG_OBJ(ctx->stt, func);

	tctx = ctx;
	if((stt_intptr_t)data & 1 && tctx->parent) tctx = tctx->parent; /* elev runs the function under the parent context */


	/* get parameter list */
	if(!stt_nextarg(ctx, &i, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for apply");
	if(i->type == STT_ST_ERROR) return i;
	if(i->type != STT_ST_CONS && i->type != STT_ST_NIL)
		return stt_create_error(ctx->stt, STT_SE_PARM, "Second parameter of apply has to be a list");
	parm = i;

	if((stt_intptr_t)data & 4) {
		/* put quotes before all arguments */
		stt_obj_t *newh = NULL, *newt = NULL;
		while(stt_nextparm(&i, &parm)) {
			if(newt) newt = newt->d.cons.cdr = STT_CONS(ctx->stt, stt_quoteobj(ctx->stt, i), STT_NIL(ctx->stt));
			else newh = newt = STT_CONS(ctx->stt, stt_quoteobj(ctx->stt, i), STT_NIL(ctx->stt));
		}
		parm = newh ? newh : STT_NIL(ctx->stt);
	}

	switch(func->type) {
	case STT_ST_FUNC:
	case STT_ST_GENERIC:
		return(stt_funcexec(tctx, parm, func, parm, tail, (stt_intptr_t)data & 2));
	case STT_ST_BUILTIN:
		return func->d.builtin.handler(tctx, parm, func->d.builtin.data, tail);
	default:
		return stt_create_error(ctx->stt, STT_SE_WFUNC, "Invalid function given to apply.");
	}
}

static STT_BUILTIN(sb_not)
{
	stt_obj_t *obj;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Not requires at least one parameter");
	if(obj->type == STT_ST_ERROR) return obj;
	if(obj->type == STT_ST_NIL)
		return stt_create_t(ctx->stt);
	else
		return stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_eq)
{
	stt_obj_t *obj, *objo;

	if(!stt_nextparm(&obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "eq requires at least one parameter");
	obj = stt_eval(ctx, obj);
	if(obj->type == STT_ST_ERROR) return obj;
	objo = obj;

	while(stt_nextparm(&obj, &parm)) {
		obj = stt_eval(ctx, obj);
		if(obj->type == STT_ST_ERROR) return obj;

		if(obj->type == STT_ST_SYMBOL) {
			if(objo->type != STT_ST_SYMBOL) return stt_create_nil(ctx->stt);
			if(objo->d.symbol.colon != obj->d.symbol.colon) return stt_create_nil(ctx->stt);
			if(objo->d.symbol.id != obj->d.symbol.id) return stt_create_nil(ctx->stt);
		} else {
			if(obj != objo) return stt_create_nil(ctx->stt);
		}
		objo = obj;
	}

	return stt_create_t(ctx->stt);
}

static STT_BUILTIN(sb_eql)
{
	stt_obj_t *obj, *o = NULL;

	if(!stt_nextarg(ctx, &o, &parm))
		stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for eql");
	if(o->type == STT_ST_ERROR) return o;

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;
		if(obj->type != o->type) return stt_create_nil(ctx->stt);

		switch(o->type) {
			case STT_ST_INT:
			case STT_ST_NUM:
			case STT_ST_STRING:
			case STT_ST_SYMBOL:
			case STT_ST_CHAR:
				if(!stt_isoform(o, obj)) return stt_create_nil(ctx->stt);
				break;
			default:
				if(o != obj) return stt_create_nil(ctx->stt);
		}
	}

	return stt_create_t(ctx->stt);
}

static STT_BUILTIN(sb_equalop)
{
	stt_obj_t *obj, *o = NULL;

	if(!stt_nextarg(ctx, &o, &parm))
		stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for equal");
	if(o->type == STT_ST_ERROR) return o;

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;

		if(!stt_isoform(o, obj)) return stt_create_nil(ctx->stt);
	}

	return stt_create_t(ctx->stt);
}

static STT_BUILTIN(sb_and)
{
	stt_obj_t *obj = NULL;

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_NIL || obj->type == STT_ST_ERROR) return obj;
	}

	return obj ? obj : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_or)
{
	stt_obj_t *obj = NULL;

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;
		if(obj->type != STT_ST_NIL) return obj;
	}

	return obj ? obj : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_cond)
{
	stt_obj_t *clause, *truth, *ret = NULL;
	while(stt_nextparm(&clause, &parm)) {
		if(clause->type != STT_ST_CONS)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Cond clause has to be a list");
		if(!stt_nextarg(ctx, &truth, &clause))
			return stt_create_error(ctx->stt, STT_SE_PARM, "Missing test for cond clause");
		if(truth->type == STT_ST_ERROR) return truth;
		if(truth->type == STT_ST_NIL) continue;

		ret = truth;
		while(stt_nextargtail(ctx, &ret, &clause, tail))
			if(ret->type == STT_ST_ERROR) return ret;
		return ret;
	}

	return stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_while)
{
	stt_obj_t *truth;
	stt_obj_t *val, *obj = NULL;
	stt_s_mark mark;

	if(!stt_nextparm(&truth, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for while");

	while(1) {
		mark = stt_gc_prot_mark(ctx->stt);
		val = stt_eval(ctx, truth);
		if(val->type == STT_ST_NIL) {
			stt_gc_prot_free(ctx->stt, mark, val);
			return val;
		}

		if(val->type == STT_ST_ERROR) {
			stt_gc_prot_free(ctx->stt, mark, val);
			if(val->d.error.code == STT_SE_RET) return val->d.error.obj ? val->d.error.obj : stt_create_nil(ctx->stt);
			return val;
		}

		val = parm;
		while(val->type == STT_ST_CONS) {
			obj = stt_eval(ctx, val->d.cons.car);
			if(obj->type == STT_ST_ERROR) {
				stt_gc_prot_free(ctx->stt, mark, obj);
				if(obj->d.error.code == STT_SE_RET) return obj->d.error.obj ? obj->d.error.obj : stt_create_nil(ctx->stt);
				return obj;
			}
			val = val->d.cons.cdr;
		}

		stt_gc_prot_free(ctx->stt, mark, NULL);
	}
}

static STT_BUILTIN(sb_return)
{
	stt_obj_t *obj, *e;
	if(!stt_nextarg(ctx, &obj, &parm))
		obj = stt_create_nil(ctx->stt);
	if(obj->type == STT_ST_ERROR) return obj;
	e = stt_create_error(ctx->stt, STT_SE_RET, NULL);
	e->d.error.obj = obj;
	return e;
}

#define sb_type_ret(x) \
	case STT_ST_ ## x: \
		return STT_SYM(ctx->stt, #x);

static STT_BUILTIN(sb_type)
{
	stt_obj_t *v;
	STT_ARG_OBJ(ctx->stt, v);
	switch(v->type) {
		sb_type_ret(NIL);
		sb_type_ret(INT);
		sb_type_ret(NUM);
		sb_type_ret(CHAR);
		sb_type_ret(T);
		sb_type_ret(STRING);
		sb_type_ret(BLOB);
		sb_type_ret(COOKIE);
		sb_type_ret(EOF);
		sb_type_ret(BUILTIN);
		sb_type_ret(ERROR);
		sb_type_ret(DEFER);
		sb_type_ret(ARRAY);
		sb_type_ret(VARCTX);
		sb_type_ret(DICT);
		sb_type_ret(STREAM);
		sb_type_ret(CLASS);
		sb_type_ret(SLOT);
		sb_type_ret(GENERIC);
		sb_type_ret(METHOD);
		sb_type_ret(DISPATCH);
		sb_type_ret(CONS);
		sb_type_ret(SYMBOL);
		sb_type_ret(VARREF);
		sb_type_ret(ANNOT);

		case STT_ST_FUNC:
			switch(v->d.func.macro) {
				case 1:
					return STT_SYM(ctx->stt, "MACRO");
				case 2:
					return STT_SYM(ctx->stt, "R-MACRO");
				case 4:
					return STT_SYM(ctx->stt, "SYMBOLIC-MACRO");
				case 5:
					return STT_SYM(ctx->stt, "R-SYMBOLIC-MACRO");
				default:
					return STT_SYM(ctx->stt, "FUNC");
			}
		case STT_ST_OBJECT:
			return v->d.clos_object.type->d.clos_class->name;
		default:
			return STT_NIL(ctx->stt);
	}
}
