static STT_BUILTIN(sb_defclass)
{
	stt_obj_t *name;
	stt_obj_t *parentlist;
	stt_obj_t *slotlist;
	stt_obj_t *cl;
	stt_var_t *var;
	stt_obj_t *iter, *iter2;
	stt_obj_t *o, *sl;

	stt_obj_t *crsr;
	stt_slotasg_t *asg;
	size_t i;

	unsigned long _ACCESSOR = stt_symbolid(ctx->stt, "ACCESSOR");
	unsigned long _WRITER = stt_symbolid(ctx->stt, "WRITER");
	unsigned long _READER = stt_symbolid(ctx->stt, "READER");
	unsigned long _ALLOCATION = stt_symbolid(ctx->stt, "ALLOCATION");
	unsigned long _INITFORM = stt_symbolid(ctx->stt, "INITFORM");
	unsigned long _INITARG = stt_symbolid(ctx->stt, "INITARG");
	unsigned long _CLASS = stt_symbolid(ctx->stt, "CLASS");
	unsigned long _INSTANCE = stt_symbolid(ctx->stt, "INSTANCE");

	/* retrieve arguments */

	if(!stt_nextparm(&name, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass needs parameter: NAME");
	if(name->type != STT_ST_SYMBOL)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: NAME must be a symbol");

	if(stt_nextparm(&parentlist, &parm)) {

		if(parentlist->type != STT_ST_NIL && parentlist->type != STT_ST_CONS)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: PARENTLIST must be a list");

		if(stt_nextparm(&slotlist, &parm)) {

			if(slotlist->type != STT_ST_NIL && slotlist->type != STT_ST_CONS)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: SLOTLIST must be a list");

		} else {
			slotlist = stt_create_nil(ctx->stt);
		}

	} else {
		parentlist = stt_create_nil(ctx->stt);
		slotlist = stt_create_nil(ctx->stt);
	}

	/* check whether class of this name already exists */

	var = stt_varctx_retr_s(ctx->stt->class_hash, name->d.symbol.sdbm, name->d.symbol.id);
	if(var) {
		return stt_create_error(ctx->stt, STT_SE_PARM, "Defclass: class redefinition isn't supported");
	}

	/* create class object */

	cl = stt_create_class(ctx->stt);
	cl->d.clos_class->typeid = stt_gettypeid(ctx->stt);
	cl->d.clos_class->name = name;
	ctx->stt->typedb[cl->d.clos_class->typeid].cl = cl;

	stt_varctx_set_s(ctx->stt->class_hash, name->d.symbol.sdbm, name->d.symbol.id, cl);

	/* look up each element in parent list and add to class object */

	crsr = NULL;
	iter = parentlist;
	while(stt_nextparm(&o, &iter)) {
		if(o->type != STT_ST_SYMBOL)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: parent name has to be a SYMBOL");

		var = stt_varctx_retr_s(ctx->stt->class_hash, o->d.symbol.sdbm, o->d.symbol.id);
		if(!var)
			return stt_create_error(ctx->stt, STT_SE_PARM, "Defclass: parent class undefined");

		if(crsr) {
			crsr->d.cons.cdr = stt_create_cons(ctx->stt, var->value, stt_create_nil(ctx->stt));
			crsr = crsr->d.cons.cdr;
		} else {
			crsr = cl->d.clos_class->parents = stt_create_cons(ctx->stt, var->value, stt_create_nil(ctx->stt));
		}
	}

	/* parse slot definitions, add them to class object */

	crsr = NULL;
	iter = slotlist;
	while(stt_nextparm(&o, &iter)) {
		sl = stt_create_slot(ctx->stt);
		sl->d.clos_slot.alloc = CL_ALLOC_INSTANCE;

		if(o->type == STT_ST_SYMBOL) {
			/* plain slot definition */
			sl->d.clos_slot.name = o;
		} else if (o->type == STT_ST_CONS) {
			/* complicated slot definition */

			iter2 = o;
			stt_nextparm(&o, &iter2);

			if(o->type != STT_ST_SYMBOL)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: slot name has to be a symbol");
			sl->d.clos_slot.name = o;

			/* read slot option list */

			while(stt_nextparm(&o, &iter2)) {
				if(o->type != STT_ST_SYMBOL)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: invalid slot definition (slot option)");

				if(o->d.symbol.id == _ACCESSOR) {

					if(!stt_nextparm(&o, &iter2))
						return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass: unexpected end of slot option list (after :ACCESSOR)");
					if(o->type != STT_ST_SYMBOL)
						return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: accessor name has to be a symbol");
					sl->d.clos_slot.accessor = o;
					continue;

				} else if(o->d.symbol.id == _WRITER) {

					if(!stt_nextparm(&o, &iter2))
						return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass: unexpected end of slot option list (after :WRITER)");
					if(o->type != STT_ST_SYMBOL)
						return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: writer name has to be a symbol");
					sl->d.clos_slot.writer = o;
					continue;

				} else if(o->d.symbol.id == _READER) {

					if(!stt_nextparm(&o, &iter2))
						return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass: unexpected end of slot option list (after :READER)");
					if(o->type != STT_ST_SYMBOL)
						return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: reader name has to be a symbol");
					sl->d.clos_slot.reader = o;
					continue;

				} else if(o->d.symbol.id == _INITARG) {

					if(!stt_nextparm(&o, &iter2))
						return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass: unexpected end of slot option list (after :INITARG)");
					if(o->type != STT_ST_SYMBOL)
						return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: initarg name has to be a symbol");
					sl->d.clos_slot.initarg = o;
					continue;

				} else if(o->d.symbol.id == _ALLOCATION) {

					if(!stt_nextparm(&o, &iter2))
						return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass: unexpected end of slot option list (after :ALLOCATION)");
					if(o->type != STT_ST_SYMBOL)
						return stt_create_error(ctx->stt, STT_SE_TYPE, "Defclass: allocation type has to be a symbol");

					if(o->d.symbol.id == _CLASS)
						sl->d.clos_slot.alloc = CL_ALLOC_CLASS;
					else if(o->d.symbol.id == _INSTANCE)
						sl->d.clos_slot.alloc = CL_ALLOC_INSTANCE;
					else
						return stt_create_error(ctx->stt, STT_SE_PARM, "Defclass: invalid allocation type");
					continue;

				} else if(o->d.symbol.id == _INITFORM) {

					if(!stt_nextparm(&o, &iter2))
						return stt_create_error(ctx->stt, STT_SE_PNUM, "Defclass: unexpected end of slot option list (after :INITFORM)");

					sl->d.clos_slot.initform = o;
					continue;

				} else {
					return stt_create_error(ctx->stt, STT_SE_PARM, "Defclass: invalid slot option");
				}

			} /* slot option list */


		} else {
			return stt_create_error(ctx->stt, STT_SE_PARM, "Defclass: invalid slot definition");
		}

		/* add ready slot object to class slot list */

		if(crsr) {
			crsr->d.cons.cdr = stt_create_cons(ctx->stt, sl, stt_create_nil(ctx->stt));
			crsr = crsr->d.cons.cdr;
		} else {
			crsr = cl->d.clos_class->slots = stt_create_cons(ctx->stt, sl, stt_create_nil(ctx->stt));
		}

	} /* slot definition list */

	/* update internal data */
	o = stt_class_calculate(ctx->stt, cl);
	if(o->type == STT_ST_ERROR) return o;

	/* evaluate initforms for static slots */

	for(i=0, asg = cl->d.clos_class->allslots; i < cl->d.clos_class->allslotsnum; i++, asg++) {
		if(asg->stor) {
			if(!cl->d.clos_class->data[asg->num]) {
				o = stt_eval(ctx, asg->slot->d.clos_slot.initform);
				if(o->type == STT_ST_ERROR) return o;
				cl->d.clos_class->data[asg->num] = o;
			}
		}
	}

	return name;
}

static STT_BUILTIN(sb_class_precedence)
{
	stt_obj_t *first = NULL, *ltail = NULL;
	stt_obj_t *o, *iter;

	iter = ctx->stt->class_precedence;
	while(stt_nextparm(&o, &iter)) {
		if(ltail) {
			ltail->d.cons.cdr = stt_create_cons(ctx->stt, o->d.clos_class->name, stt_create_nil(ctx->stt));
			ltail = ltail->d.cons.cdr;
		} else {
			first = ltail = stt_create_cons(ctx->stt, o->d.clos_class->name, stt_create_nil(ctx->stt));
		}
	}

	return first ? first : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_class_ancestry)
{
	stt_var_t *var;
	stt_obj_t *iter, *obj;
	stt_obj_t *first = NULL, *ltail = NULL;

	STT_ARG_SYMBOL(ctx->stt, obj);

	var = stt_varctx_retr_s(ctx->stt->class_hash, obj->d.symbol.sdbm, obj->d.symbol.id);
	if(!var)
		return stt_create_error(ctx->stt, STT_SE_UNDEF, "Class-ancestry: undefined class");

	iter = var->value->d.clos_class->ancestry;
	while(stt_nextparm(&obj, &iter)) {
		if(ltail) {
			ltail->d.cons.cdr = stt_create_cons(ctx->stt, obj->d.clos_class->name, stt_create_nil(ctx->stt));
			ltail = ltail->d.cons.cdr;
		} else {
			first = ltail = stt_create_cons(ctx->stt, obj->d.clos_class->name, stt_create_nil(ctx->stt));
		}
	}

	return first ? first : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_class_slotinfo)
{
	stt_var_t *var;
	size_t i;
	stt_obj_t *first = NULL, *ltail = NULL;
	stt_obj_t *info, *obj;
	stt_slotasg_t *asg;

	STT_ARG_SYMBOL(ctx->stt, obj);

	var = stt_varctx_retr_s(ctx->stt->class_hash, obj->d.symbol.sdbm, obj->d.symbol.id);
	if(!var)
		return stt_create_error(ctx->stt, STT_SE_UNDEF, "Class-slotinfo: undefined class");

	obj=var->value;

	for(i=0; i<obj->d.clos_class->allslotsnum; i++) {
		asg = obj->d.clos_class->allslots + i;
		info = stt_create_nil(ctx->stt);
		if(asg->slot->d.clos_slot.accessor)
			info = stt_create_cons(ctx->stt, stt_create_cons(ctx->stt, stt_create_keyword(ctx->stt, "ACCESSOR"), asg->slot->d.clos_slot.accessor), info);
		if(asg->slot->d.clos_slot.writer)
			info = stt_create_cons(ctx->stt, stt_create_cons(ctx->stt, stt_create_keyword(ctx->stt, "WRITER"), asg->slot->d.clos_slot.writer), info);
		if(asg->slot->d.clos_slot.reader)
			info = stt_create_cons(ctx->stt, stt_create_cons(ctx->stt, stt_create_keyword(ctx->stt, "READER"), asg->slot->d.clos_slot.reader), info);

		info = stt_create_cons(ctx->stt, asg->slot->d.clos_slot.name, info);

		if(ltail) {
			ltail->d.cons.cdr = stt_create_cons(ctx->stt, info, stt_create_nil(ctx->stt));
			ltail = ltail->d.cons.cdr;
		} else {
			first = ltail = stt_create_cons(ctx->stt, info, stt_create_nil(ctx->stt));
		}
	}

	return first ? first : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_slot_value)
{
	stt_obj_t *type, *obj, *slot;
	stt_var_t *var;
	stt_slotasg_t *asg;

	STT_ARG_OBJECT(ctx->stt, obj);
	STT_ARG_SYMBOL(ctx->stt, slot);

	type = obj->d.clos_object.type;
	var = stt_varctx_retr_s(type->d.clos_class->slotname, slot->d.symbol.sdbm, slot->d.symbol.id);
	if(!var)
		return stt_create_error(ctx->stt, STT_SE_UNDEF, "Slot-value: undefined slot");

	asg = type->d.clos_class->allslots + var->value->d.integer.value;

	if(asg->stor) {
		return asg->stor->d.clos_class->data[asg->num];
	} else {
		return obj->d.clos_object.data[asg->num];
	}
}

static STT_BUILTIN(sb_slot_value_set)
{
	stt_obj_t *type, *obj, *slot, *value;
	stt_var_t *var;
	stt_slotasg_t *asg;

	STT_ARG_OBJECT(ctx->stt, obj);
	STT_ARG_SYMBOL(ctx->stt, slot);
	STT_ARG_OBJ(ctx->stt, value);

	type = obj->d.clos_object.type;
	var = stt_varctx_retr_s(type->d.clos_class->slotname, slot->d.symbol.sdbm, slot->d.symbol.id);
	if(!var)
		return stt_create_error(ctx->stt, STT_SE_UNDEF, "Slot-value-set: undefined slot");

	asg = type->d.clos_class->allslots + var->value->d.integer.value;

	if(asg->stor) {
		asg->stor->d.clos_class->data[asg->num] = value;
	} else {
		obj->d.clos_object.data[asg->num] = value;
	}

	return value;
}

static STT_BUILTIN(sb_slot_value_set_weak)
{
	stt_obj_t *type, *obj, *slot, *value;
	stt_var_t *var;
	stt_slotasg_t *asg;
	stt_obj_t **pntr;

	STT_ARG_OBJECT(ctx->stt, obj);
	STT_ARG_SYMBOL(ctx->stt, slot);
	STT_ARG_OBJ(ctx->stt, value);

	type = obj->d.clos_object.type;
	var = stt_varctx_retr_s(type->d.clos_class->slotname, slot->d.symbol.sdbm, slot->d.symbol.id);
	if(!var)
		return stt_create_error(ctx->stt, STT_SE_UNDEF, "Slot-value-set-weak: undefined slot");

	asg = type->d.clos_class->allslots + var->value->d.integer.value;

	if(asg->stor) {
		pntr = asg->stor->d.clos_class->data + asg->num;
		*pntr = value;
		stt_weak_check(value);
		stt_weak_reference(asg->stor, asg->stor->serial, pntr);
	} else {
		pntr = obj->d.clos_object.data + asg->num;
		*pntr = value;
		stt_weak_check(value);
		stt_weak_reference(obj, obj->serial, pntr);
	}

	return value;
}

static STT_BUILTIN(sb_make_instance)
{
	stt_obj_t *type;
	stt_obj_t *inst;
	stt_var_t *var;
	stt_obj_t *name, *value;
	stt_slotasg_t *asg;
	size_t i;

	if(!stt_nextarg(ctx, &type, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for make-instance");
	if(type->type == STT_ST_ERROR) return type;
	if(type->type != STT_ST_SYMBOL)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "make-instance: type mismatch (not symbol)");

	var = stt_varctx_retr_s(ctx->stt->class_hash, type->d.symbol.sdbm, type->d.symbol.id);
	if(!var)
		return stt_create_error(ctx->stt, STT_SE_UNDEF, "make-instance: undefined class");
	type = var->value;

	inst = stt_create_object(ctx->stt, type);

	/* parse initargs */

	while(stt_nextarg(ctx, &name, &parm)) {
		if(name->type == STT_ST_ERROR) return name;
		if(name->type != STT_ST_SYMBOL)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "make-instance: initarg name must be a symbol");

		if(!stt_nextarg(ctx, &value, &parm))
			return stt_create_error(ctx->stt, STT_SE_PNUM, "make-instance: missing value after initarg name");
		if(value->type == STT_ST_ERROR) return value;

		var = stt_varctx_retr_s(type->d.clos_class->initarg, name->d.symbol.sdbm, name->d.symbol.id);
		if(!var)
			return stt_create_error(ctx->stt, STT_SE_UNDEF, "make-instance: initarg undefined");

		asg = type->d.clos_class->allslots + var->value->d.integer.value;

		if(asg->stor) {
			asg->stor->d.clos_class->data[asg->num] = value;
		} else {
			inst->d.clos_object.data[asg->num] = value;
		}
	}

	/* write defaults/NILs to unset slots */

	for(i=0, asg = type->d.clos_class->allslots; i < type->d.clos_class->allslotsnum; i++, asg++) {
		if(!asg->stor) {
			if(!inst->d.clos_object.data[asg->num]) {
				value = stt_eval(ctx, asg->slot->d.clos_slot.initform);
				if(value->type == STT_ST_ERROR) return value;
				inst->d.clos_object.data[asg->num] = value;
			}
		}
	}

	return inst;
}

static STT_BUILTIN(sb_is_generalisation)
{
	stt_obj_t *class_a, *class_b;
	stt_var_t *vara, *varb;

	STT_ARG_SYMBOL(ctx->stt, class_a);
	STT_ARG_SYMBOL(ctx->stt, class_b);

	vara = stt_varctx_retr_s(ctx->stt->class_hash, class_a->d.symbol.sdbm, class_a->d.symbol.id);
	if(!vara) return stt_create_error(ctx->stt, STT_SE_PARM, "Class A undefined");
	varb = stt_varctx_retr_s(ctx->stt->class_hash, class_b->d.symbol.sdbm, class_b->d.symbol.id);
	if(!varb) return stt_create_error(ctx->stt, STT_SE_PARM, "Class B undefined");

	return stt_is_generalisation(vara->value, varb->value) ? stt_create_t(ctx->stt) : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_call_next_method)
{
	stt_var_t * var = stt_varctx_retr_s(ctx, ctx->stt->dispatch_info_sdbm, ctx->stt->dispatch_info_id);
	stt_obj_t *dispatch, *item, *oldm, *rtrn;
	int i = 0;

	if(!var) return stt_create_error(ctx->stt, STT_SE_UNDEF, "call-next-method called outside a dispatch");
	dispatch = var->value;
	if(dispatch->type != STT_ST_DISPATCH) return stt_create_error(ctx->stt, STT_SE_TYPE, "_DISPATCH-TYPE not a dispatch.");

	while(stt_nextarg(ctx, &item, &parm)) {
		if(item->type == STT_ST_ERROR) return item;
		if(i == 0) {
			/* if there are any arguments, create new dispatch info */
			stt_obj_t *nd;
			nd = stt_create_dispatch(ctx->stt);
			nd->d.clos_dispatch.method = dispatch->d.clos_dispatch.method;
			nd->d.clos_dispatch.generic = dispatch->d.clos_dispatch.generic;
			dispatch = nd;
		}

		dispatch->d.clos_dispatch.argeval[i++] = item;
	}
	if(i > 0)
		dispatch->d.clos_dispatch.argeval[i] = NULL;

	oldm = dispatch->d.clos_dispatch.method;

	rtrn = stt_funcexec(ctx, oldm, dispatch, stt_create_nil(ctx->stt), 0, 0);

	dispatch->d.clos_dispatch.method = oldm;

	return rtrn;
}

static STT_BUILTIN(sb_next_method_p)
{
	stt_var_t * var = stt_varctx_retr_s(ctx, ctx->stt->dispatch_info_sdbm, ctx->stt->dispatch_info_id);
	stt_obj_t *dispatch, *method;

	if(!var) return stt_create_error(ctx->stt, STT_SE_UNDEF, "next-method-p called outside a dispatch");
	dispatch = var->value;
	if(dispatch->type != STT_ST_DISPATCH) return stt_create_error(ctx->stt, STT_SE_TYPE, "_DISPATCH-TYPE not a dispatch.");

	method = dispatch->d.clos_dispatch.method;

	for(;;) {
		method = method->d.clos_method.next;

		if(!method) return stt_create_nil(ctx->stt);
		if(stt_is_applicable(ctx->stt, method, dispatch->d.clos_dispatch.argeval)) return stt_create_t(ctx->stt);
	}
}
