static STT_BUILTIN(sb_eval)
{
	stt_obj_t *obj, *j = NULL;

	while(stt_nextarg(ctx, &obj, &parm)) {
		j = stt_eval(ctx, obj); /* evaluate twice */
		if(j->type == STT_ST_ERROR)
			return j;
	}

	return j ? j : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_trace)
{
	stt_obj_t *obj;
	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for trace");
	if(obj->type == STT_ST_ERROR) return obj;
	if(obj->type != STT_ST_FUNC)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "trace requires a function argument");

	obj->d.func.trace = (stt_intptr_t) data;
	return obj;
}

static STT_BUILTIN(sb_gc)
{
	int swept;

	/* make (gc) do a full sweep */
	ctx->stt->gc_gen2_count = 1;
	ctx->stt->gc_gen3_count = 1;

	swept = stt_gc_sweep(ctx->stt, ctx);
	return STT_LIST2(ctx->stt, STT_INT(ctx->stt, swept), STT_INT(ctx->stt, ctx->stt->gc_live));
}

static STT_BUILTIN(sb_dumpctx)
{
	return stt_varctx_dump(ctx);
}

static STT_BUILTIN(sb_limit)
{
	stt_obj_t *obj;
	long lim;

	if(!stt_next_t(ctx, &obj, &parm, STT_ST_INT))
		return stt_create_error(ctx->stt, STT_SE_PARM, "Invalid parameters for limit");
	lim = ctx->stt->recursion_limit;
	ctx->stt->recursion_limit = obj->d.integer.value;
	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) {
			ctx->stt->recursion_limit = lim;
			return obj;
		}
	}

	ctx->stt->recursion_limit = lim;
	return obj;
}

static char loadnmbuf[512];
static char loadbuf[512];

static STT_BUILTIN(sb_load_path)
{
	return STT_STR(ctx->stt, ctx->stt->load_path);
}

static STT_BUILTIN(sb_load)
{
	stt_obj_t *obj = NULL;
	char *st, *nam;
	struct stat stb;
	int l;

	if(!((stt_intptr_t) data & 1))
		while(ctx->parent) ctx = ctx->parent; /* load globally */

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;
		if(obj->type != STT_ST_STRING && obj->type != STT_ST_SYMBOL)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "load: Type mismatch (not string or symbol)");

		/* check len of string to avoid buffer overflows.. */
		nam = st = (obj->type == STT_ST_STRING ? obj->d.string.value : obj->d.symbol.name);
		l = strlen(st);
		if(l>256) return stt_create_error(ctx->stt, STT_SE_PARM, "Too long include >_<");

		if(obj->type == STT_ST_SYMBOL) {
			strcpy(loadnmbuf, st);
			strcat(loadnmbuf, ".stt");
			nam = st = loadnmbuf;
		}

		if((stt_intptr_t) data & 2) { /* called as 'require' */
			const char *st, *e;
			st = ctx->stt->refpath.str;
			while(*st) {
				e = strchr(st, STT_LOAD_SEPCHAR);
				if(!e) {
					strcpy(loadbuf, st);
				} else {
					strncpy(loadbuf, st, (e - st));
					loadbuf[e - st] = 0;
				}
				strcat(loadbuf, STT_PATHSEP);
				strcat(loadbuf, nam);

				if(!stat(loadbuf, &stb)) {
					obj = stt_evalfile(ctx, loadbuf);
					if(obj && obj->type == STT_ST_ERROR) return obj;
					goto nextload;
				}

				if(e) st = e + 1; else break;
			}

			return stt_create_error_with_obj(ctx->stt, STT_SE_NFND, "Not found", stt_create_string(ctx->stt, nam));
		} else {
			obj = stt_evalfile(ctx, st);
			if(obj && obj->type == STT_ST_ERROR) return obj;
		}

nextload:;
	}

	return obj ? obj : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_throw)
{
	stt_obj_t *eobj = NULL, *obj;
	long code;
	char *msg = NULL;

	STT_ARG_INT(ctx->stt, obj);

	code = obj->d.integer.value;

	if(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;
		if(obj->type != STT_ST_STRING)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Error message needs to be a string");
		msg = obj->d.string.value;

		if(stt_nextarg(ctx, &obj, &parm)) {
			if(obj->type == STT_ST_ERROR) return obj;
			eobj = obj;
		}
	}

	obj = stt_create_error(ctx->stt, code, msg);
	obj->d.error.obj = eobj;

	if((stt_intptr_t)data == 1) {
		obj->d.error.file = ctx->stt->file;
		obj->d.error.line = ctx->stt->parsed_line;
	}

	return obj;
}
