static STT_BUILTIN(sb_get_refpath)
{
	return STT_STR(ctx->stt, ctx->stt->refpath.str);
}

static STT_BUILTIN(sb_set_refpath)
{
	char *refp;
	stt_obj_t *stt_refp;
	STT_L_STRING_OBJ(ctx->stt, refp);

	ctx->stt->refpath.len = 0;
	stt_strbuild_addstr(&ctx->stt->refpath, refp, strlen(refp));
	return stt_refp;
}

static STT_BUILTIN(sb_print)
{
	stt_obj_t *obj = NULL;
	stt_obj_t *o;
	int f = 0;
	while(stt_nextarg(ctx, &obj, &parm)) {
		if((stt_intptr_t) data != 2) {
			if(f) {
				o = stt_stream_stdout_write(ctx, ' ');
				if(o->type == STT_ST_ERROR) return obj;
			}
		}

		if(obj->type == STT_ST_ERROR) return obj;

		if((stt_intptr_t) data) { /* princ/princ- */
			if(obj->type == STT_ST_STRING) {
				o = stt_stream_dumpbuf(ctx, (unsigned char *) obj->d.string.value, strlen(obj->d.string.value));
			} else if(obj->type == STT_ST_CHAR) {
				o = stt_stream_stdout_write(ctx, obj->d.ch.ch);
			} else {
				o = stt_printobj(ctx, obj);
			}
		} else { /* print */
			o = stt_printobj(ctx, obj);
		}

		if(o->type == STT_ST_ERROR) return o;

		f = 1;
	}

	return obj ? obj : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_println)
{
	stt_obj_t *obj;
	stt_obj_t *o;
	obj = sb_print(ctx, parm, data, tail);
	if(obj->type == STT_ST_ERROR) return obj;
	o = stt_stream_stdout_write(ctx, '\n');
	if(o->type == STT_ST_ERROR) return o;
	return obj;
}

static STT_BUILTIN(sb_read_char)
{
	stt_obj_t *str;

	if(stt_nextarg(ctx, &str, &parm)) {
		if(str->type == STT_ST_ERROR) return str;
		if(str->type != STT_ST_STREAM) return stt_create_error(ctx->stt, STT_SE_TYPE, "read-char: Type mismatch (not stream)");
		return stt_stream_read_char(ctx, str);
	}

	return stt_stream_stdin_read(ctx);
}

static STT_BUILTIN(sb_unread_char)
{
	stt_obj_t *str, *chr;
	STT_ARG_OBJ(ctx->stt, str);
	STT_ARG_CHAR(ctx->stt, chr);

	return stt_stream_unread_char(ctx->stt, str, chr->d.ch.ch);
}

static STT_BUILTIN(sb_read_chunk)
{
	stt_obj_t *stream;
	long bytes;
	char *buf;

	STT_ARG_STREAM(ctx->stt, stream);
	STT_L_INT(ctx->stt, bytes);

	if(bytes < 1)
		return stt_create_error(ctx->stt, STT_SE_IO, "read-chunk: bytes argument must not be less than 1");

	buf = (char *) stt_malloc(bytes + 1);

	if(stream->d.stream.type == STT_STREAM_FP) {
		int r = fread(buf, 1, bytes, stream->d.stream.s.fp);
		if(r == 0) {
			free(buf);
			return stt_create_eof(ctx->stt);
		}

		buf[r] = 0;
		return stt_create_string_allocd(ctx->stt, buf);
	} else if(stream->d.stream.type == STT_STREAM_C && stream->d.stream.s.c.readmore) {
		int r = stream->d.stream.s.c.readmore(&stream->d.stream.s.c.data, stream->d.stream.s.c.type, buf, bytes);
		if(r == -1) {
			free(buf);
			return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "read-chunk: I/O error in readmore", STT_STR(ctx->stt, strerror(errno)));
		}
		if(r == 0) {
			free(buf);
			return stt_create_eof(ctx->stt);
		}
		buf[r] = 0;
		return stt_create_string_allocd(ctx->stt, buf);
	} else {
		int l = 0;
		for(;;) {
			stt_obj_t *o = stt_stream_read_char(ctx, stream);

			if(o->type == STT_ST_ERROR) {
				free(buf);
				return o;
			}

			if(o->type == STT_ST_EOF) {
				if(l > 0) {
					buf[l] = 0;
					return stt_create_string_allocd(ctx->stt, buf);
				}

				free(buf);
				return stt_create_eof(ctx->stt);
			}

			buf[l++] = o->d.ch.ch;
			if(l == bytes) {
				buf[l] = 0;
				return stt_create_string_allocd(ctx->stt, buf);
			}
		}
	}
}

static STT_BUILTIN(sb_read_line)
{
	stt_obj_t *str;
	stt_obj_t *stream = NULL;
	stt_obj_t *o;
	int le = 1, l = 0;
	char *out;
	str = stt_create_string(ctx->stt, "");

	if(stt_nextarg(ctx, &stream, &parm)) {
		if(stream->type == STT_ST_ERROR) return stream;
		if(stream->type != STT_ST_STREAM) return stt_create_error(ctx->stt, STT_SE_TYPE, "read-line takes a stream parameter");
	}

	if(!stream) stream = stt_get_input_stream(ctx);
	if(stream->type == STT_ST_ERROR) return stream;

	if(stream->d.stream.type == STT_STREAM_FP) {
		/* optimization for FP based streams */
		char buf[512];
		int sl;
		if(!fgets(buf, 512, stream->d.stream.s.fp)) {
			int err = errno;
			if(feof(stream->d.stream.s.fp))
				return stt_create_eof(ctx->stt);
			return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "IO error in FP optimized read-line", STT_STR(ctx->stt, strerror(err)));
		}

		sl = strlen(buf);
		if(buf[sl-1] == '\n') {
			buf[sl-1] = 0;
			return stt_create_string(ctx->stt, buf);
		}

		l = sl;
		out = (char *) malloc(l+1);
		if(!out) return stt_create_error(ctx->stt, STT_SE_MEM, "Out of memory");
		strcpy(out, buf);

		while(fgets(buf, 512, stream->d.stream.s.fp)) {
			sl = strlen(buf);
			out = (char *) realloc(out, l + sl + 1);
			if(!out) return stt_create_error(ctx->stt, STT_SE_MEM, "Out of memory");
			if(buf[sl-1] == '\n') {
				buf[sl-1] = 0;
				strcpy(out + l, buf);
				return stt_create_string_allocd(ctx->stt, out);
			}

			assert(sl == 511);
			strcpy(out + l, buf);
			l += sl;
		}
		if(!feof(stream->d.stream.s.fp))
			return stt_create_error(ctx->stt, STT_SE_IO, "IO error in FP optimized read-line");
		return stt_create_string_allocd(ctx->stt, out);
	}

	/* fallback: unoptimized streams */
	while(1) {
		o = stt_stream_read_char(ctx, stream);

		if(o->type == STT_ST_ERROR)
			return o;
		if(o->type == STT_ST_EOF) {
			if(l > 0) {
				str->d.string.value[l] = 0;
				break;
			} else return o;
		}

		if(le == l + 1) {
			le = le * 2 + 1;
			str->d.string.value = realloc(str->d.string.value, le);
			if(!str->d.string.value) return stt_create_error(ctx->stt, STT_SE_MEM, "Out of memory");
		}

		str->d.string.value[l++] = o->d.ch.ch;
		if(o->d.ch.ch == '\n') {
			str->d.string.value[l-1] = 0;
			break;
		}
	}

	return str;
}

static STT_BUILTIN(sb_read)
{
	int ol;
	stt_read_state_t *rs;
	stt_obj_t *o;
	stt_obj_t *stream = NULL;
	int c;

	if(stt_nextarg(ctx, &stream, &parm)) {
		if(stream->type == STT_ST_ERROR) return stream;
		if(stream->type != STT_ST_STREAM) return stt_create_error(ctx->stt, STT_SE_TYPE, "read-line takes a stream parameter");
	}

	if(!stream) stream = stt_get_input_stream(ctx);
	if(stream->type == STT_ST_ERROR) return stream;

	ol = ctx->stt->parsed_line;
	rs = stt_create_read_state(ctx->stt);

	for(;;) {
		o = stt_stream_read_char(ctx, stream);

		if(o->type == STT_ST_ERROR) {
			stt_free_read_state(rs);
			ctx->stt->parsed_line = ol;
			return o;
		}
		if(o->type == STT_ST_EOF)
			c = EOF;
		else
			c = o->d.ch.ch;

		o = stt_read_ch(rs, c);
		if(o && o->type == STT_ST_ERROR) {
			stt_free_read_state(rs);
			ctx->stt->parsed_line = ol;
			return o;
		}

		if(o || c == EOF) break;
	}

	stt_free_read_state(rs);
	ctx->stt->parsed_line = ol;

	if(o) return o;
	return stt_create_eof(ctx->stt);
}

static STT_BUILTIN(sb_make_stream)
{
	stt_obj_t *obj;
	stt_obj_t *str;

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "make-stream requires a parameter");
	if(obj->type == STT_ST_ERROR) return obj;
	if(obj->type != STT_ST_FUNC || obj->d.func.macro > 0)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "make-stream requires a lambda parameter");

	str = stt_create_stream(ctx->stt, STT_STREAM_FUNC);
	str->d.stream.s.func = obj;
	return str;
}

static STT_BUILTIN(sb_open_file)
{
	stt_obj_t *mode, *obj;
	FILE *fp;

	STT_ARG_STRING(ctx->stt, obj);

	if(!stt_nextarg(ctx, &mode, &parm)) {
		fp = fopen(obj->d.string.value, "r");
	} else {
		if(mode->type == STT_ST_ERROR) return mode;
		if(mode->type != STT_ST_STRING)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "open-file file mode has to be a string");

		fp = fopen(obj->d.string.value, mode->d.string.value);
	}

	if(!fp)
		return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "Cannot open file", stt_create_string(ctx->stt, strerror(errno)));
	obj = stt_create_stream(ctx->stt, STT_STREAM_FP);
	obj->d.stream.s.fp = fp;
	return obj;
}

static STT_BUILTIN(sb_dump_file)
{
	char *fname, *buf;
	long r, skip = 0, len = -1;
	stt_obj_t *seek;
	FILE *fp;

	STT_L_STRING(ctx->stt, fname);

	if(stt_nextarg(ctx, &seek, &parm)) {
		if(seek->type == STT_ST_ERROR) return seek;
		if(seek->type != STT_ST_INT)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "dump-file seek has to be integer");

		skip = seek->d.integer.value;

		if(stt_nextarg(ctx, &seek, &parm)) {
			if(seek->type == STT_ST_ERROR) return seek;
			if(seek->type != STT_ST_INT)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "dump-file len has to be integer");
			len = seek->d.integer.value;
			if(len < 1)
				return stt_create_error(ctx->stt, STT_SE_PARM, "dump-file len has to be > 0");
		}
	}

	fp = fopen(fname, "r");
	if(!fp)
		return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "Cannot open file", stt_create_string(ctx->stt, strerror(errno)));

	if(fseek(fp, skip, SEEK_SET) == -1) {
		fclose(fp);
		return stt_create_error(ctx->stt, STT_SE_IO, "dump-file seek fails");
	}

	buf = (char *) stt_malloc(4096);
	if(len == -1) {
		/* unlimited read */
		while((r = fread(buf, 1, 4096, fp)) > 0) {
			stt_obj_t *o = stt_stream_dumpbuf(ctx, (unsigned char *) buf, r);
			if(o->type == STT_ST_ERROR) {
				free(buf);
				fclose(fp);
				return o;
			}
		}
	} else {
		/* limited read */
		int togo = len;
		stt_obj_t *o;
		while(togo > 0) {
			if(feof(fp)) {
				fclose(fp);
				return stt_create_error(ctx->stt, STT_SE_IO, "dump-file: premature end of file");
			}

			r = fread(buf, 1, togo > 4096 ? 4096 : togo, fp);
			if(r < 1) break;
			o = stt_stream_dumpbuf(ctx, (unsigned char *) buf, r);
			if(o->type == STT_ST_ERROR) {
				free(buf);
				fclose(fp);
				return o;
			}
			togo -= r;
		}
	}

	free(buf);
	fclose(fp);
	return STT_NIL(ctx->stt);
}

static STT_BUILTIN(sb_fflush)
{
	stt_obj_t *obj;
	STT_ARG_STREAM(ctx->stt, obj);

	if(obj->d.stream.type != STT_STREAM_FP)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Invalid stream (not opened with open-file?)");
	if(!obj->d.stream.s.fp)
		return stt_create_error(ctx->stt, STT_SE_IO, "Stream already closed");

	return fflush(obj->d.stream.s.fp) ? stt_create_nil(ctx->stt) : stt_create_t(ctx->stt);
}

static STT_BUILTIN(sb_close_file)
{
	stt_obj_t *obj;
	STT_ARG_STREAM(ctx->stt, obj);

	if(obj->d.stream.type != STT_STREAM_FP)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Invalid stream (not opened with open-file?)");
	if(!obj->d.stream.s.fp)
		return stt_create_error(ctx->stt, STT_SE_IO, "Stream already closed");
	if(fclose(obj->d.stream.s.fp))
		return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "Stream close failed", stt_create_string(ctx->stt, strerror(errno)));
	obj->d.stream.s.fp = NULL;

	return stt_create_t(ctx->stt);
}

/* stream writer for string buckets */
static size_t bucket_writemore_(char **data, const int type, void *buf, size_t len)
{
	if(!*data) {
		*data = stt_calloc(1, len + 1 + 2 * sizeof(size_t));
		*((size_t *) *data) = len + 1;
		*(((size_t *) *data) + 1) = len + 1;
		memcpy(*data + 2 * sizeof(size_t), buf, len);
		((char *) (*data)) [(2 * sizeof(size_t)) + len] = 0;
	} else {
		size_t b = *((size_t *) *data);
		size_t l = *(((size_t *) *data) + 1);

		while(l + len > b) b *= 2;
		*data = stt_realloc(*data, b + 2 * sizeof(size_t));
		*((size_t *) *data) = b;
		*(((size_t *) *data) + 1) = l + len;
		memcpy(*data + 2 * sizeof(size_t) + l - 1, buf, len);
		((char *) (*data)) [(2 * sizeof(size_t)) + len + l - 1] = 0;
	}

	return 0;
}

static size_t bucket_writemore(void **data, const int type, void *buf, size_t len)
{
	return bucket_writemore_((char **)data, type, buf, len); /* work around void ** -> char ** */
}

static size_t bucket_vprintf(void **data, const int type, const char *fmt, va_list ap_orig)
{
	char buf[2048], *dst;
	int need2free;
	size_t r, ret;

	dst = stt_asvprintf(buf, sizeof(buf), &need2free, &r, fmt, ap_orig);

	ret = bucket_writemore(data, type, (unsigned char *)dst, r);
	if (need2free)
		free(dst);
	return ret;
}

/* stream writer for string buckets */
static int bucket_writer(int ch, void **data, const int type)
{
	size_t buf, len;

	if(!*data) {
		*data = stt_calloc(1, 8 + 2 * sizeof(size_t));
		buf = *((size_t *) *data) = 8;
		len = *(((size_t *) *data) + 1) = 1;
	} else {
		buf = *((size_t *) *data);
		len = *(((size_t *) *data) + 1);
	}

	if(len == buf) {
		buf *= 2;
		*data = stt_realloc(*data, buf + 2 * sizeof(size_t));
		*((size_t *) *data) = buf;
	}

	((char *) (*data)) [(2 * sizeof(size_t)) + len] = 0;
	((char *) (*data)) [(2 * sizeof(size_t)) + ((len++) - 1)] = ch;

	*(((size_t *) *data) + 1) = len;

	return 0;
}

static void string_bucket_destroy(stt_obj_t *obj)
{
	free(obj->d.stream.s.c.data);
}

static STT_BUILTIN(sb_string_bucket)
{
	stt_obj_t *stream = stt_create_stream(ctx->stt, STT_STREAM_C);
	stream->d.stream.s.c.writer = bucket_writer;
	stream->d.stream.s.c.writemore = bucket_writemore;
	stream->d.stream.s.c.vprintf = bucket_vprintf;
	stream->d.stream.s.c.destroy = string_bucket_destroy;
	return stream;
}

static STT_BUILTIN(sb_empty_bucket)
{
	stt_obj_t *str, *obj, *blobp;
	int asblob = 0;

	STT_ARG_STREAM(ctx->stt, obj);

	if(parm->type == STT_ST_CONS) {
		STT_ARG_OBJ(ctx->stt, blobp);
		asblob = !(blobp->type == STT_ST_NIL);
	}

	if(obj->d.stream.type != STT_STREAM_C || obj->d.stream.s.c.type != 0)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Invalid type of stream");

	if(!obj->d.stream.s.c.data)
		return asblob ? stt_create_blob(ctx->stt, NULL, 0) : stt_create_string(ctx->stt, "");

	if(asblob) {
		size_t len = *(((size_t *) obj->d.stream.s.c.data) + 1);
		len --;
		return stt_create_blob_copy(ctx->stt, (unsigned char *) ((char **)obj->d.stream.s.c.data + 2 * sizeof(size_t)), len);
	} else {
		str = stt_create_string(ctx->stt, (char *) ((char **)obj->d.stream.s.c.data + 2*sizeof(size_t)));
	}

	free(obj->d.stream.s.c.data);
	obj->d.stream.s.c.data = NULL;
	return str;
}

/* stream reader for string-stream */
static int string_stream_reader(void **data, const int type)
{
	char **x = (char **) *data;
	int c = **x;
	if(!c) return -1;
	(*x) ++;
	return c;
}

/* stream reader for string-stream */
static size_t string_stream_readmore(void **data, const int type, void *buf, const size_t len)
{
	size_t l;
	char **x = (char **) *data;
	char *d = buf;
	for(l=0; l<len; l++) {
		*d = **x;
		if(!*d) return l;
		d++;
		(*x) ++;
	}
	return l;
}

/* string-stream destructor */
static void string_stream_destroy(stt_obj_t *obj)
{
	free(obj->d.stream.s.c.data);
}

static STT_BUILTIN(sb_string_stream)
{
	char *instr;
	stt_obj_t *stream;
	size_t l;
	char *d;

	STT_L_STRING(ctx->stt, instr);
	stream = stt_create_stream(ctx->stt, STT_STREAM_C);
	l = strlen(instr);
	d = stt_malloc(l + 1 + sizeof(char *));
	strcpy(*((char **) d) = d + sizeof(char *), instr);
	stream->d.stream.s.c.data = d;
	stream->d.stream.s.c.reader = string_stream_reader;
	stream->d.stream.s.c.readmore = string_stream_readmore;
	stream->d.stream.s.c.destroy = string_stream_destroy;
	return stream;
}

static STT_BUILTIN(sb_system)
{
	stt_obj_t *command;
	STT_ARG_STRING(ctx->stt, command);

#ifdef WIN32
	return stt_create_int(ctx->stt, system(command->d.string.value));
#else
	return stt_create_int(ctx->stt, WEXITSTATUS(system(command->d.string.value)));
#endif
}
