static STT_BUILTIN(sb_rename)
{
	stt_obj_t *oldpath, *newpath;

	if (!stt_nextarg(ctx, &oldpath, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "RENAME requires two parameters");
	if (oldpath->type == STT_ST_ERROR)
		return oldpath;
	if (!stt_nextarg(ctx, &newpath, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "RENAME requires two parameters");
	if (newpath->type == STT_ST_ERROR)
		return newpath;

	if (oldpath->type != STT_ST_STRING)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "RENAME's first parameter must be a string");
	if (newpath->type != STT_ST_STRING)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "RENAME's second parameter must be a string");

	if (rename(oldpath->d.string.value, newpath->d.string.value) < 0)
		return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "Cannot rename file",
			stt_create_string(ctx->stt, strerror(errno)));

	return stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_unlink)
{
	stt_obj_t *path;

	if (!stt_nextarg(ctx, &path, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "UNLINK requires a parameter");
	if (path->type == STT_ST_ERROR)
		return path;

	if (path->type != STT_ST_STRING)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "UNLINK's first parameter must be a string");

	if (remove(path->d.string.value) < 0)
		return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "Cannot unlink file",
			stt_create_string(ctx->stt, strerror(errno)));

	return stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_getenv)
{
	stt_obj_t *name;
	char *v;
	STT_ARG_STRING(ctx->stt, name);

	v = getenv(name->d.string.value);
	return v ? stt_create_string(ctx->stt, v) : stt_create_nil(ctx->stt);
}

static stt_obj_t *make_fileinfo(stt_varctx_t *ctx, struct stat *sb)
{
	long size[1] = {3};
	stt_obj_t *ar = stt_create_array(ctx->stt, 1, size);
	if(S_ISREG(sb->st_mode))
		ar->d.array.data[0] = stt_create_symbol(ctx->stt, "REG");
	else if(S_ISDIR(sb->st_mode))
		ar->d.array.data[0] = stt_create_symbol(ctx->stt, "DIR");
	else
		ar->d.array.data[0] = stt_create_nil(ctx->stt);

	ar->d.array.data[1] = stt_create_int(ctx->stt, sb->st_size);
	ar->d.array.data[2] = stt_create_int(ctx->stt, sb->st_mtime);
	return stt_create_cons(ctx->stt, stt_create_symbol(ctx->stt, "FILE"), ar);
}

static STT_BUILTIN(sb_file)
{
	stt_obj_t *path;
	struct stat sb;
	int res;

	STT_ARG_STRING(ctx->stt, path);
	res = stat(path->d.string.value, &sb);
	if(res) return stt_create_error_with_obj(ctx->stt, STT_SE_IO, "file: can't stat", path);
	return make_fileinfo(ctx, &sb);
}
