STT_BUILTIN(stt_sb_quote)
{
	if(parm->type != STT_ST_CONS)
		return stt_create_error(ctx->stt, STT_SE_PARM, "QUOTE needs a parameter");

	return parm->d.cons.car;
}

static STT_BUILTIN(sb_lambda)
{
	stt_obj_t *prm, *exp = NULL, *first, *cur;
	stt_obj_t *docstr = NULL;

	if(!stt_nextparm(&prm, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for lambda");
	if(!prm || (prm->type != STT_ST_CONS && prm->type != STT_ST_NIL))
		return stt_create_error(ctx->stt, STT_SE_PARM, "First parameter of lambda must be a list of argument names");

	first = cur = NULL;
	while(stt_nextparm(&exp, &parm)) {
		if(!cur) {
			if(!docstr && exp->type == STT_ST_STRING) docstr = exp;
			else
				first = cur = stt_create_cons(ctx->stt, exp, stt_create_nil(ctx->stt));
		} else {
			cur->d.cons.cdr = stt_create_cons(ctx->stt, exp, stt_create_nil(ctx->stt));
			cur = cur->d.cons.cdr;
		}
	}

	if(!first && docstr) {
		first = stt_create_cons(ctx->stt, docstr, stt_create_nil(ctx->stt));
		docstr = NULL;
	}

	exp = first ? first : stt_create_nil(ctx->stt);
	cur = stt_create_func(ctx->stt, prm, exp);
	cur->d.func.docstr = docstr;
	cur->d.func.macro = ((stt_intptr_t) data) & 7;
		/* 1 for macro, 2 for r-macro
		 * 4 for symbolic-macro, 5 for r-symbolic-macro */

	if(ctx->parent && (stt_intptr_t) data != 8) {
		/* stt_varctx_t isn't global.. create a closure */
		if(!ctx->ref) {
			cur->d.func.closure = stt_create_obj_varctx(ctx->stt, ctx);
			ctx->ref = cur->d.func.closure;
		} else cur->d.func.closure = ctx->ref;

		/* keep all parents */
		ctx = ctx->parent;
		while(ctx->parent) {
			if(!ctx->ref) ctx->ref = stt_create_obj_varctx(ctx->stt, ctx);
			ctx = ctx->parent;
		}
	}

	return cur;
}

static STT_BUILTIN(sb_if)
{
	stt_obj_t *obj;
	stt_obj_t *val;
	stt_obj_t *tr;

	if(!stt_nextparm(&obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for if");
	val = stt_eval(ctx, obj);
	if(val->type == STT_ST_ERROR) return val;

	if(!stt_nextparm(&tr, &parm))
		return stt_create_error(ctx->stt, STT_SE_PARM, "Insufficient parameters for if");

	if(val->type == STT_ST_NIL) {
		/* false */
		if(!stt_nextparm(&tr, &parm))
			return stt_create_nil(ctx->stt);
		else
			return stt_eval_(ctx, tr, tail);
	} else {
		/* true */
		return stt_eval_(ctx, tr, tail);
	}
}

static STT_BUILTIN(sb_subctx)
{
	stt_obj_t *obj, *j = NULL;
	stt_varctx_t *fctx;

	fctx = stt_varctx_create(ctx->stt, ctx, 64);

	while(stt_nextparm(&obj, &parm)) {
		j = stt_eval(fctx, obj);
		if(j->type == STT_ST_ERROR) {
			if(!fctx->ref) stt_varctx_free(fctx);
			return j;
		}
	}

	if(!fctx->ref) stt_varctx_free(fctx);

	return j ? j : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_do)
{
	stt_obj_t *bind, *test = NULL, *asg, *i, *obj, *var, *ret = NULL;
	stt_s_mark mark;
	stt_obj_t *vals[FUNC_MAXPARMS];
	stt_varctx_t *nctx;
	int let = (stt_intptr_t) data;
	int nasg;

	if(!stt_nextparm(&bind, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for do/let");
	if(bind->type != STT_ST_CONS && bind->type != STT_ST_NIL)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "First argument of do/let has to be a list");
	if(!let) {
		if(!stt_nextparm(&test, &parm))
			return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for do/let");
		if(test->type != STT_ST_CONS)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Second argument of do has to be a list");
	}

	nctx = stt_varctx_create(ctx->stt, ctx, FUNC_LOCALCTX_SIZE);
	nctx->weak = 1;

	/* assign initial values */

	i = bind; nasg = 0;
	while(stt_nextparm(&asg, &i)) {
		if(nasg == FUNC_MAXPARMS) {
			if(!nctx->ref) stt_varctx_free(nctx);
			return stt_create_error(ctx->stt, STT_SE_PNUM, "Too many assignment clauses in do/let");
		}

		if(asg->type != STT_ST_CONS) {
			if(!nctx->ref) stt_varctx_free(nctx);
			return stt_create_error(ctx->stt, STT_SE_TYPE, "First argument of do/let has to be a list of lists");
		}

		stt_nextparm(&var, &asg);
		if(var->type != STT_ST_SYMBOL) {
			if(!nctx->ref) stt_varctx_free(nctx);
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Do/let can't assign to a non-symbol");
		}

		if(!stt_nextarg(ctx, &obj, &asg)) {
			obj = stt_create_nil(ctx->stt);
		} else if(obj->type == STT_ST_ERROR) {
			if(!nctx->ref) stt_varctx_free(nctx);
			if(!let && obj->d.error.code == STT_SE_RET) return obj->d.error.obj ? obj->d.error.obj : stt_create_nil(ctx->stt);
			return obj;
		}

		stt_varctx_store_s(nctx, var->d.symbol.sdbm, var->d.symbol.id, obj);
		nasg ++;
	}

	/* the loop */

	while(1) {
		mark = stt_gc_prot_mark(ctx->stt);

		if(!let) { /* (do), test condition */
			i = test;
			stt_nextparm(&ret, &i);
			ret = stt_eval(nctx, ret);
			if(ret->type == STT_ST_ERROR) {
				if(!nctx->ref) stt_varctx_free(nctx);
				if(!let && ret->d.error.code == STT_SE_RET) return ret->d.error.obj ? ret->d.error.obj : stt_create_nil(ctx->stt);
				return ret;
			}
			if(ret->type != STT_ST_NIL) {
				if(stt_nextarg(nctx, &obj, &i)) { /* return value specified */
					if(!nctx->ref) stt_varctx_free(nctx);
					return obj;
				}

				if(!nctx->ref) stt_varctx_free(nctx); /* return value not specified, returning test result */
				return ret;
			}
		}

		/* execute body */
		i = parm;
		while(i->type == STT_ST_CONS) {
			ret = stt_eval(nctx, i->d.cons.car);
			if(ret->type == STT_ST_ERROR) {
				if(!nctx->ref) stt_varctx_free(nctx);
				if(!let && ret->d.error.code == STT_SE_RET) return ret->d.error.obj ? ret->d.error.obj : stt_create_nil(ctx->stt);
				return ret;
			}
			i = i->d.cons.cdr;
		}

		if(let) break; else {
			/* execute adjustments */

			i = bind;
			nasg = 0;
			while(stt_nextparm(&asg, &i)) { /* evaluate values */
				vals[nasg] = stt_varctx_retr_s(nctx, asg->d.cons.car->d.symbol.sdbm, asg->d.cons.car->d.symbol.id)->value;
				asg = asg->d.cons.cdr;
				if(asg->type != STT_ST_CONS) {
					nasg++; continue;
				}

				asg = asg->d.cons.cdr;
				if(asg->type == STT_ST_CONS) {
					asg = asg->d.cons.car;
					obj = stt_eval(nctx, asg);
					if(obj->type == STT_ST_ERROR) {
						if(!nctx->ref) stt_varctx_free(nctx);
						if(!let && obj->d.error.code == STT_SE_RET) return obj->d.error.obj ? obj->d.error.obj : stt_create_nil(ctx->stt);
						return obj;
					}
					vals[nasg] = obj;
				}
				nasg++;
			}

			/* assign values */
			nasg = 0; i = bind;
			while(stt_nextparm(&asg, &i)) {
				stt_varctx_set_s(nctx, asg->d.cons.car->d.symbol.sdbm, asg->d.cons.car->d.symbol.id, vals[nasg]);
				nasg ++;
			}

			stt_gc_prot_free(ctx->stt, mark, NULL);
		}
	}

	if(!nctx->ref) stt_varctx_free(nctx);

	return ret ? ret : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_catch)
{
	stt_obj_t *list, *car, *func;
	stt_obj_t *msgo;
	stt_obj_t *obj = NULL;
	int cgt;

	if(!stt_nextarg(ctx, &list, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for catch");
	if(list->type == STT_ST_ERROR) return list;
	if(list->type != STT_ST_CONS && list->type != STT_ST_T)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Error codes parameter for catch must be a list or T");

	/* check list for types */
	if(list->type == STT_ST_CONS) {
		for(car = list; car->type == STT_ST_CONS; car = car->d.cons.cdr) {
			if(car->d.cons.car->type != STT_ST_INT)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "Error codes must be integers");
		}
	}

	STT_ARG_FUNCTION(ctx->stt, func);

	while(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) {
			/* it's an error, look in the list whether we should catch it */
			cgt = 0;
			if(list->type == STT_ST_T) {
				cgt = 1;
			} else {
				for(car = list; car->type == STT_ST_CONS; car = car->d.cons.cdr) {
					if(car->d.cons.car->d.integer.value == obj->d.error.code) {
						cgt = 1;
						break;
					}
				}
			}

			if(cgt) {
				/* caught it, call handler */
				if(obj->d.error.msg) msgo = stt_create_string(ctx->stt, obj->d.error.msg); else msgo = stt_create_nil(ctx->stt);
				func = STT_LIST5(ctx->stt, func,
						STT_INT(ctx->stt, obj->d.error.code),
						msgo,
						obj->d.error.obj ? stt_quoteobj(ctx->stt, obj->d.error.obj) : stt_create_nil(ctx->stt),
						stt_quoteobj(ctx->stt, STT_CONS(ctx->stt, obj, STT_NIL(ctx->stt))));

				obj = stt_eval(ctx, func);
				break;
			}

			return obj;
		}
	}

	return obj ? obj : stt_create_nil(ctx->stt);
}

static STT_BUILTIN(sb_defmethod)
{
	stt_obj_t *name;
	stt_obj_t *place, *args = NULL;
	stt_obj_t *code;
	stt_obj_t *method;
	stt_obj_t *iter;
	stt_obj_t *arg, *res;
	int argn;

	int mplace = CL_METHOD_PRIMARY;

	unsigned long _BEFORE = stt_symbolid(ctx->stt, "BEFORE");
	unsigned long _PRIMARY = stt_symbolid(ctx->stt, "PRIMARY");
	unsigned long _AFTER = stt_symbolid(ctx->stt, "AFTER");
	unsigned long _AROUND = stt_symbolid(ctx->stt, "AROUND");
	unsigned long _PLUS = stt_symbolid(ctx->stt, "+");
	unsigned long _EQL = stt_symbolid(ctx->stt, "EQL");

	/* grab method name */
	if(!stt_nextparm(&name, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Defmethod needs parameter: NAME");
	if(name->type != STT_ST_SYMBOL)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "Defmethod: NAME is expected to be a symbol");

	/* next argument is either method place or the arglist */
	if(!stt_nextparm(&place, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Defmethod needs parameter: ARGLIST");
	if(place->type == STT_ST_SYMBOL) {
		/* method place it is */

		if(place->d.symbol.id == _BEFORE)
			mplace = CL_METHOD_BEFORE;
		else if(place->d.symbol.id == _PRIMARY)
			mplace = CL_METHOD_PRIMARY;
		else if(place->d.symbol.id == _AFTER)
			mplace = CL_METHOD_AFTER;
		else if(place->d.symbol.id == _AROUND)
			mplace = CL_METHOD_AROUND;
		else
			return stt_create_error(ctx->stt, STT_SE_PARM, "Defmethod: invalid method place");

		/* grab arglist after this */
		if(!stt_nextparm(&args, &parm))
			return stt_create_error(ctx->stt, STT_SE_PNUM, "Defmethod needs parameter: ARGLIST (after PLACE)");
		if(args->type != STT_ST_CONS && args->type != STT_ST_NIL)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "Defmethod: ARGLIST must be a list");

	} else if(place->type == STT_ST_CONS || place->type == STT_ST_NIL) {
		/* arglist it is */
		args = place;
	} else return stt_create_error(ctx->stt, STT_SE_TYPE, "Defmethod: NAME must be followed by a PLACE or an ARGLIST");

	/* everything after this is code for the method */
	code = parm;

	method = stt_create_method(ctx->stt);
	method->d.clos_method.place = mplace;
	method->d.clos_method.code = code;

	/* count number of arguments */
	argn = 0;
	iter = args;
	while(stt_nextparm(&arg, &iter)) {
		/* break after :+
		 * (the following arguments aren't considered for dispatch) */
		if(arg->type == STT_ST_SYMBOL && arg->d.symbol.id == _PLUS && arg->d.symbol.colon) break;
		argn ++;
	}

	if(argn) {
		/* make place for the argument structure */
		method->d.clos_method.args = (stt_method_arg_t *) stt_calloc(argn, sizeof(stt_method_arg_t));
		method->d.clos_method.argn = argn;

		/* parse arguments */
		argn = 0;
		iter = args;

		while(stt_nextparm(&arg, &iter)) {
			if(arg->type == STT_ST_SYMBOL) {
				/* end of significant arglist? */
				if(arg->d.symbol.id == _PLUS && arg->d.symbol.colon)
					break;

				/* simple untyped arg */
				method->d.clos_method.args[argn].name = arg;
				method->d.clos_method.args[argn].type = -1;
			} else if(arg->type == STT_ST_CONS) {
				/* some kind of classified arg. */

				stt_obj_t *iter2 = arg;
				stt_obj_t *item;

				stt_nextparm(&item, &iter2);
				if(item->type != STT_ST_SYMBOL)
					return stt_create_error(ctx->stt, STT_SE_TYPE, "Defmethod: argument name in classified pair must be a symbol");

				method->d.clos_method.args[argn].name = item;

				if(stt_nextparm(&item, &iter2)) {
					/* classification */

					if(item->type == STT_ST_SYMBOL) {
						/* should be a type */

						int typeid = stt_symbol_to_typeid(ctx->stt, item);
						if(typeid == -1)
							return stt_create_error_with_obj(ctx->stt, STT_SE_PARM, "Defmethod: invalid type (in classification)", item);
						method->d.clos_method.args[argn].type = typeid;
					} else if(item->type == STT_ST_CONS) {
						/* (eql something)? */

						stt_obj_t *iter3 = item;
						stt_obj_t *elem;

						stt_nextparm(&elem, &iter3);
						if(elem->type != STT_ST_SYMBOL)
							return stt_create_error(ctx->stt, STT_SE_PARM, "Defmethod: symbol expected as classification type list's first element");
						if(elem->d.symbol.id != _EQL)
							return stt_create_error(ctx->stt, STT_SE_PARM, "Defmethod: only EQL is supported as classification type magic");

						stt_nextparm(&elem, &iter3);
						elem = stt_eval(ctx, elem);
						if(elem->type == STT_ST_ERROR) return elem;

						method->d.clos_method.args[argn].type = stt_object_typeid(elem);
						method->d.clos_method.args[argn].value = elem;
					} else {
						/* not a symbol, not a cons. */
						return stt_create_error(ctx->stt, STT_SE_PARM, "Defmethod: type in classified pair must be a symbol or a list");
					}
				}

			} else {
				/* argument name not a symbol, not a cons. */
				return stt_create_error(ctx->stt, STT_SE_PARM, "Defmethod: argument name must be a symbol or a list");
			}

			argn ++;
		}

		/* set misc args to rest of arglist */
		method->d.clos_method.miscargs = iter;
	} else {
		/* argn == 0 */
		method->d.clos_method.miscargs = args;
	}

	res = stt_method_calculate(ctx, name, method);
	if(res->type == STT_ST_ERROR) return res;

	return name;
}

