static STT_BUILTIN(sb_toupper)
{
	stt_obj_t *obj;
	char *c;
	int (*conv)(int) = (int (*)(int))stt_V2F(data);

	STT_ARG_STRING(ctx->stt, obj);

	obj = stt_create_string(ctx->stt, obj->d.string.value); /* this stt_strdups */
	for(c = obj->d.string.value; *c; c++) {
		*c = conv(*c);
	}

	return obj;
}

static STT_BUILTIN(sb_utf8_length)
{
	stt_obj_t *str;
	unsigned char *c;
	unsigned long int code = 0;
	long len = 0;
	int pending = 0;

	STT_ARG_STRING(ctx->stt, str);

	for(c = (unsigned char *) str->d.string.value; *c; c++) {
		if (*c <= 0x7E) {
			if (pending != 0)
				goto BAD_UTF8;
			code = *c;
		} else if (*c <= 0xBF) {
			if (pending == 0)
				goto BAD_UTF8;
			pending--;
			code <<= 6;
			code |= *c & 0x3F;
		} else if (*c <= 0xDF) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 1;
			code = *c & 0x1F;
		} else if (*c <= 0xEF) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 2;
			code = *c & 0x0F;
		} else if (*c <= 0xF7) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 3;
			code = *c & 0x07;
		} else if (*c <= 0xFB) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 4;
			code = *c & 0x03;
		} else if (*c <= 0xFD) {
			if (pending != 0)
				goto BAD_UTF8;
			pending = 5;
			code = *c & 0x01;
		} else {
			goto BAD_UTF8;
		}
		if (pending == 0) {
			len ++;
		}
	}
	if(pending) {
BAD_UTF8:
		return stt_create_error(ctx->stt, STT_SE_PARM, "UTF8-LENGTH's parameter must be a UTF-8 encoded string");
	}
	return stt_create_int(ctx->stt, len);
}

static STT_BUILTIN(sb_substr)
{
	stt_obj_t *obj;
	char *str, *nstr;
	size_t start = 0, len = 0, le;
	size_t lenunset = 1;

	STT_ARG_STRING(ctx->stt, obj);

	str = obj->d.string.value;
	le = strlen(str);

	if(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;
		if(obj->type != STT_ST_INT)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "substr: Type mismatch (start not int)");
		start = obj->d.integer.value;

		if(stt_nextarg(ctx, &obj, &parm)) {
			if(obj->type == STT_ST_ERROR) return obj;
			if(obj->type != STT_ST_INT)
				return stt_create_error(ctx->stt, STT_SE_TYPE, "substr: Type mismatch (len not int)");
			len = obj->d.integer.value;
			lenunset = 0;
		}
	}

	if(!lenunset && !len) return stt_create_string(ctx->stt, "");

	le = strlen(str);
	if(!le) return stt_create_string(ctx->stt, "");

	if(start < 0) {
		start = le - (0 - start) % le;
		if(start == le) start = 0;
	}

	if(start >= le) return stt_create_string(ctx->stt, "");

	if(lenunset) len = le;
	if(len > (le - start)) len = le - start;
	if(len < 0) len = le + len - start;
	if(len < 0) return stt_create_string(ctx->stt, "");

	nstr = (char *) stt_malloc(len + 1);
	strncpy(nstr, str + start, len);
	*(nstr + len) = 0;

	return stt_create_string_allocd(ctx->stt, nstr);
}

static STT_BUILTIN(sb_explode)
{
	stt_obj_t *obj;
	char *str, *needle, *p;
	stt_obj_t *list = NULL, *ltail = NULL;
	size_t limit = -1, lneedle;

	STT_ARG_STRING(ctx->stt, obj);

	needle = obj->d.string.value;
	lneedle = strlen(needle);
	if(!lneedle) return stt_create_nil(ctx->stt);

	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for explode");
	if(obj->type == STT_ST_ERROR) return obj;
	if(obj->type != STT_ST_STRING)
		return stt_create_error(ctx->stt, STT_SE_TYPE, "explode: Type mismatch (str not string)");
	str = obj->d.string.value;

	if(stt_nextarg(ctx, &obj, &parm)) {
		if(obj->type == STT_ST_ERROR) return obj;
		if(obj->type != STT_ST_INT)
			return stt_create_error(ctx->stt, STT_SE_TYPE, "explode: Type mismatch (limit not int)");
		limit = obj->d.integer.value;
	}

	for(;;) {
		if(!*str) {
			if(!ltail) list = ltail = stt_create_cons(ctx->stt, stt_create_string(ctx->stt, ""), stt_create_nil(ctx->stt));
			else {
				ltail->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_string(ctx->stt, ""), stt_create_nil(ctx->stt));
				ltail = ltail->d.cons.cdr;
			}
			return list;
		}

		p = strstr(str, needle);

		limit--;
		if(limit == 0 || !p) {
			if(!ltail) list = ltail = stt_create_cons(ctx->stt, stt_create_string(ctx->stt, str), stt_create_nil(ctx->stt));
			else ltail->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_string(ctx->stt, str), stt_create_nil(ctx->stt));
			return list;
		}

		if(!ltail) list = ltail = stt_create_cons(ctx->stt, stt_create_string_counted(ctx->stt, str, p - str), stt_create_nil(ctx->stt));
		else {
			ltail->d.cons.cdr = stt_create_cons(ctx->stt, stt_create_string_counted(ctx->stt, str, p - str), stt_create_nil(ctx->stt));
			ltail = ltail->d.cons.cdr;
		}

		str = p + lneedle;
	}
}

static STT_BUILTIN(sb_replace)
{
	stt_obj_t *stt_haystack;
	char *needle, *haystack, *repl, *at, *p;
	size_t ln, nln, rln;
	stt_strbuild_t bldr;

	STT_L_STRING(ctx->stt, needle);
	STT_L_STRING_OBJ(ctx->stt, haystack);
	STT_L_STRING(ctx->stt, repl);

	bldr = stt_strbuild_new();

	at = haystack;
	ln = strlen(haystack);
	nln = strlen(needle);
	rln = strlen(repl);

	if(nln == 0) return stt_haystack;

	while((p = strstr(at, needle))) {
		if(p > at)
			stt_strbuild_addstr(&bldr, at, (p - at));

		stt_strbuild_addstr(&bldr, repl, rln);

		at = p + nln;
	}

	if(at < haystack + ln)
		stt_strbuild_addstr(&bldr, at, (ln - (at - haystack)));

	if(bldr.str)
		return stt_create_string_allocd(ctx->stt, bldr.str);
	else
		return stt_create_string(ctx->stt, "");
}

static STT_BUILTIN(sb_trim)
{
	stt_obj_t *str;
	int l, start, end, i, j;
	char *bf;

	STT_ARG_STRING(ctx->stt, str);
	l = strlen(str->d.string.value);
	bf = (char *) stt_malloc(l + 1);

	for(start=0; start<l; start++) {
		if(!isspace(str->d.string.value[start])) break;
	}

	for(end=l-1; end>=0; end--) {
		if(!isspace(str->d.string.value[end])) break;
	}

	if(end == -1) {
		start = 0;
	}

	for(i = start; i<=end; i++) {
		bf[j++] = str->d.string.value[i];
	}
	bf[j] = 0;

	return stt_create_string_allocd(ctx->stt, bf);
}

static STT_BUILTIN(sb_char_code)
{
	stt_obj_t *obj;
	if(!stt_nextarg(ctx, &obj, &parm))
		return stt_create_error(ctx->stt, STT_SE_PNUM, "Insufficient parameters for char-code");
	if(obj->type == STT_ST_ERROR) return obj;
	switch(obj->type) {
		case STT_ST_STRING:
			return stt_create_int(ctx->stt, (unsigned char) obj->d.string.value[0]);
		case STT_ST_CHAR:
			return stt_create_int(ctx->stt, (unsigned char) obj->d.ch.ch);
		default:
			return stt_create_error(ctx->stt, STT_SE_TYPE, "char->code: Type mismatch (not string)");
	}
}

static STT_BUILTIN(sb_code_char)
{
	stt_obj_t *obj;
	STT_ARG_INT(ctx->stt, obj);
	return stt_create_char(ctx->stt, obj->d.integer.value);
}

static STT_BUILTIN(sb_nthchar)
{
	size_t i, l;
	stt_obj_t *obj, *str;

	STT_ARG_INT(ctx->stt, obj);
	STT_ARG_OBJ(ctx->stt, str);

	i = obj->d.integer.value;

	if(str->type == STT_ST_STRING)
	{
		l = strlen(str->d.string.value);
		if(!l) return stt_create_error(ctx->stt, STT_SE_PARM, "Empty string given to nthchar");

		if(i > 0) i = i % l;
		else if(i < 0) i = l - ((0-i) % l);

		return stt_create_char(ctx->stt, str->d.string.value[i]);
	} else if(str->type == STT_ST_BLOB) {
		l = str->d.blob.size;
		if(!l) return stt_create_error(ctx->stt, STT_SE_PARM, "Empty blob given to nthchar");

		if(i > 0) i = i % l;
		else if(i < 0) i = l - ((0-i) % l);

		return stt_create_char(ctx->stt, str->d.blob.data[i]);
	} else return stt_create_error(ctx->stt, STT_SE_TYPE, "nthchar: parameter not string or blob");
}

