#ifndef _BUILTINS_H
#define _BUILTINS_H

#include <estutter/types.h>
#include <estutter/memory.h>
#include <estutter/stt_config.h>

#define STT_PATHSEP "/"
#define STT_LOAD_SEPCHAR ';'
#define STT_LOAD_SEPSTR ";"

/* handy macros for builtin definition
 * STT_ARG_X(stt, name)
 *  defines stt_obj_t *name
 *  retrieves next argument from argument list
 *  checks for error and type
 * STT_GRAB_X(stt, name)
 *  as STT_ARG_X, but doesn't define variable (reuses existing)
 */

#define STT_BUILTIN(name) stt_obj_t * name (stt_varctx_t *ctx, stt_obj_t *parm, void *data, const int tail)

#define STT_GRAB_OBJ(stt, name) \
	if(!stt_nextarg(ctx, & name, &parm)) \
		return stt_create_error(stt, STT_SE_PNUM, "Insufficient arguments, missing " #name); \
	if(name->type == STT_ST_ERROR) return name;

#define STT_ARG_OBJ(stt, name) STT_GRAB_OBJ(stt, name)

#define STT_ARG_OBJ_T(stt, name, dtype) \
	STT_ARG_OBJ(stt, name); \
	if(name->type != dtype) return stt_create_error(stt, STT_SE_TYPE, "Type mismatch: " #name " expected to be " #dtype);

#define STT_GRAB_OBJ_T(stt, name, dtype) \
	STT_GRAB_OBJ(stt, name); \
	if(name->type != dtype) return stt_create_error(stt, STT_SE_TYPE, "Type mismatch: " #name " expected to be " #dtype);

#define STT_ARG_FUNCTION(stt, name) \
	STT_ARG_OBJ(stt, name); \
	if(name->type != STT_ST_BUILTIN && name->type != STT_ST_FUNC) \
		return stt_create_error(stt, STT_SE_TYPE, "Type mismatch: " #name " expected to be a function");

#define STT_ARG_LIST(stt, name) \
	STT_ARG_OBJ(stt, name); \
	if(name->type != STT_ST_NIL && name->type != STT_ST_CONS) \
		return stt_create_error(stt, STT_SE_TYPE, "Type mismatch: " #name " expected to be a list");

#define STT_GRAB_FUNCTION(stt, name) \
	STT_GRAB_OBJ(stt, name); \
	if(name->type != STT_ST_BUILTIN && name->type != STT_ST_FUNC) \
		return stt_create_error(stt, STT_SE_TYPE, "Type mismatch: " #name " expected to be a function");

#define STT_GRAB_LIST(stt, name) \
	STT_GRAB_OBJ(stt, name); \
	if(name->type != STT_ST_NIL && name->type != STT_ST_CONS) \
		return stt_create_error(stt, STT_SE_TYPE, "Type mismatch: " #name " expected to be a list");

#define STT_ARG_INT(stt, name)     STT_ARG_OBJ_T(stt, name, STT_ST_INT);
#define STT_ARG_NUM(stt, name)     STT_ARG_OBJ_T(stt, name, STT_ST_NUM);
#define STT_ARG_CHAR(stt, name)    STT_ARG_OBJ_T(stt, name, STT_ST_CHAR);
#define STT_ARG_STRING(stt, name)  STT_ARG_OBJ_T(stt, name, STT_ST_STRING);
#define STT_ARG_ARRAY(stt, name)   STT_ARG_OBJ_T(stt, name, STT_ST_ARRAY);
#define STT_ARG_DICT(stt, name)    STT_ARG_OBJ_T(stt, name, STT_ST_DICT);
#define STT_ARG_STREAM(stt, name)  STT_ARG_OBJ_T(stt, name, STT_ST_STREAM);
#define STT_ARG_CLASS(stt, name)   STT_ARG_OBJ_T(stt, name, STT_ST_CLASS);
#define STT_ARG_SLOT(stt, name)    STT_ARG_OBJ_T(stt, name, STT_ST_SLOT);
#define STT_ARG_OBJECT(stt, name)  STT_ARG_OBJ_T(stt, name, STT_ST_OBJECT);
#define STT_ARG_GENERIC(stt, name) STT_ARG_OBJ_T(stt, name, STT_ST_GENERIC);
#define STT_ARG_METHOD(stt, name)  STT_ARG_OBJ_T(stt, name, STT_ST_METHOD);
#define STT_ARG_SYMBOL(stt, name)  STT_ARG_OBJ_T(stt, name, STT_ST_SYMBOL);
#define STT_ARG_CONS(stt, name)    STT_ARG_OBJ_T(stt, name, STT_ST_CONS);
#define STT_ARG_BLOB(stt, name)    STT_ARG_OBJ_T(stt, name, STT_ST_BLOB);
#define STT_ARG_COOKIE(stt, name, cookiet) \
	STT_ARG_OBJ_T(stt, name, STT_ST_COOKIE); \
	if(name->d.cookie.type != cookiet) return stt_create_error(stt, STT_SE_PARM, "Expected cookie type: " #cookiet);

#define STT_GRAB_INT(stt, name)     STT_GRAB_OBJ_T(stt, name, STT_ST_INT);
#define STT_GRAB_NUM(stt, name)     STT_GRAB_OBJ_T(stt, name, STT_ST_NUM);
#define STT_GRAB_CHAR(stt, name)    STT_GRAB_OBJ_T(stt, name, STT_ST_CHAR);
#define STT_GRAB_STRING(stt, name)  STT_GRAB_OBJ_T(stt, name, STT_ST_STRING);
#define STT_GRAB_ARRAY(stt, name)   STT_GRAB_OBJ_T(stt, name, STT_ST_ARRAY);
#define STT_GRAB_DICT(stt, name)    STT_GRAB_OBJ_T(stt, name, STT_ST_DICT);
#define STT_GRAB_STREAM(stt, name)  STT_GRAB_OBJ_T(stt, name, STT_ST_STREAM);
#define STT_GRAB_CLASS(stt, name)   STT_GRAB_OBJ_T(stt, name, STT_ST_CLASS);
#define STT_GRAB_SLOT(stt, name)    STT_GRAB_OBJ_T(stt, name, STT_ST_SLOT);
#define STT_GRAB_OBJECT(stt, name)  STT_GRAB_OBJ_T(stt, name, STT_ST_OBJECT);
#define STT_GRAB_GENERIC(stt, name) STT_GRAB_OBJ_T(stt, name, STT_ST_GENERIC);
#define STT_GRAB_METHOD(stt, name)  STT_GRAB_OBJ_T(stt, name, STT_ST_METHOD);
#define STT_GRAB_SYMBOL(stt, name)  STT_GRAB_OBJ_T(stt, name, STT_ST_SYMBOL);
#define STT_GRAB_CONS(stt, name)    STT_GRAB_OBJ_T(stt, name, STT_ST_CONS);
#define STT_GRAB_BLOB(stt, name)    STT_GRAB_OBJ_T(stt, name, STT_ST_BLOB);
#define STT_GRAB_COOKIE(stt, name, cookiet) \
	STT_GRAB_OBJ_T(stt, name, STT_ST_COOKIE); \
	if(name->d.cookie.type != cookiet) return stt_create_error(stt, STT_SE_PARM, "Expected cookie type: " #cookiet);

#define STT_CONS(stt, a,b) stt_create_cons((stt), (a), (b))
#define STT_NIL(stt) stt_create_nil((stt))
#define STT_INT(stt, n) stt_create_int((stt), (n))
#define STT_NUM(stt, n) stt_create_num((stt), (n))
#define STT_SYM(stt, s) stt_create_symbol((stt), (s))
#define STT_STR(stt, s) stt_create_string((stt), (s))

#define STT_CAR(o) ((o)->d.cons.car)
#define STT_CDR(o) ((o)->d.cons.cdr)
#define STT_SECOND(o) STT_CAR(STT_CDR(o))
#define STT_THIRD(o) STT_CAR(STT_CDR(CDR(o)))
#define STT_FOURTH(o) STT_CAR(STT_CDR(CDR(STT_CDR(o))))
#define STT_FIFTH(o) STT_CAR(STT_CDR(CDR(STT_CDR(CDR(o)))))
#define STT_SIXTH(o) STT_CAR(STT_CDR(CDR(STT_CDR(CDR(STT_CDR(o))))))

#define STT_LIST2(stt,a,b) STT_CONS((stt), (a), STT_CONS((stt), (b), STT_NIL(stt)))
#define STT_LIST3(stt,a,b,c) STT_CONS((stt), (a), STT_CONS((stt), (b), STT_CONS((stt), (c), STT_NIL(stt))))
#define STT_LIST4(stt,a,b,c,d) STT_CONS((stt), (a), STT_CONS((stt), (b), STT_CONS((stt), (c), STT_CONS((stt), (d), STT_NIL(stt)))))
#define STT_LIST5(stt,a,b,c,d,e) STT_CONS((stt), (a), STT_LIST4((stt), (b), (c), (d), (e)))
#define STT_LIST6(stt,a,b,c,d,e,f) STT_CONS((stt), (a), STT_LIST5((stt), (b), (c), (d), (e), (f)))

#define STT_SSYM(stt, s) \
	static stt_obj_t * s = NULL; \
	if(! s) s = stt_create_symbol(stt, #s);

#define STT_L_INT(stt, n) \
	{ \
		stt_obj_t *__tmp__; \
		STT_ARG_INT(stt, __tmp__); \
		n = __tmp__->d.integer.value; \
	}

#define STT_L_STRING(stt, n) \
	{ \
		stt_obj_t *__tmp__; \
		STT_ARG_STRING(stt, __tmp__); \
		n = __tmp__->d.string.value; \
	}

#define STT_L_STRING_OBJ(stt, n) \
	STT_ARG_STRING(stt, stt_ ## n); \
	n = stt_ ## n->d.string.value;

#define STT_L_NUM(stt, n) \
	{ \
		stt_obj_t *__tmp__; \
		STT_ARG_NUM(stt, __tmp__); \
		n = __tmp__->d.num.value; \
	}

/* print cons-dump */
stt_obj_t *stt_consdump(stt_varctx_t *ctx, stt_obj_t *obj);

/* get value of next parameter (returns 1 on success, 0 if there are no more) */
STT_INLINE int stt_nextparm(stt_obj_t **obj, stt_obj_t **parm);

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more) */
STT_INLINE int stt_nextarg(stt_varctx_t *ctx, stt_obj_t **obj, stt_obj_t **parm);

/* get and evaluate next parameter (returns 1 on success, 0 if there are no more).
 * if it's the last parameter and tail is true, calls eval
 * with tail flag set to 1! */
STT_INLINE int stt_nextargtail(stt_varctx_t *ctx, stt_obj_t **obj, stt_obj_t **parm, const int tail);

/* get and evaluate next parameter, ensure it's from a type
 * returns 1 on success */
int stt_next_t(stt_varctx_t *ctx, stt_obj_t **obj, stt_obj_t **parm, const int type);

/* register builtins in variable context */
void stt_register_builtins(stt_varctx_t *ctx);

/* register stdio as streams in variable context */
void stt_register_stdio_as(stt_varctx_t *ctx, FILE *in, FILE *out, FILE *err);
void stt_register_stdio(stt_varctx_t *ctx); /* automatic for global stdio */

#endif
