#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "compat.h"

char *stt_strdup(const char *src)
{
	int l = strlen(src);
	char *dst;
	dst = malloc(l+1);
	memcpy(dst, src, l+1);
	return dst;
}


int stt_strcasecmp(const char *str1, const char *str2)
{
	while(tolower(*str1) == tolower(*str2)) {
		if (*str1 == '\0') return 0;
		str1++;
		str2++;
	}
	return tolower(*str1) - tolower(*str2);
}

char *stt_asvprintf(char *small, size_t small_len, int *need2free, size_t *rlen, const char *fmt, va_list ap_orig)
{
#ifdef va_copy
	/* C99 has va_copy and most probably a working vsnprintf() as well. Most
	   systems are C99 by now. This method uses stack allocation for small
	   buffers and grows (by doubling) a large buffer on heap when needed */
	va_list ap;
	char *dst;
	int ret, r, l, f = 0;

	va_copy(ap, ap_orig);
	r = vsnprintf(small, small_len, fmt, ap);
	if(r >= 0 && r < small_len) {
		va_end(ap);
		*rlen = r;
		*need2free = 0;
		return small;
	} else {
		va_end(ap);
		f = 1;   /* denote |dst| needs to be |free|d */
		l = small_len*2; /* initial buffer size */
		while(1) {
			dst = (char *) stt_malloc(l);
			va_copy(ap, ap_orig);
			r = vsnprintf(dst, l, fmt, ap);
			va_end(ap);
			if(r >= 0 && r < l)
				break;
			free(dst);
			l *= 2;
		}
	}

	*rlen = r;
	*need2free = 1;
	return dst;
#else
	/* On old systems we need to go for one large buffer and risk data loss. */
	size_t r, bs = 65536;
	char *buf = malloc(bs);
	r = vsprintf(buf, fmt, ap_orig);
	if (r >= bs-1) {
		fprintf(stderr, "stt_asprintf: pre-c99 system can not print more than %ld bytes at once.\n", (long)bs-1);
		abort(); /* the heap is already broken. */
	}
	*rlen = r;
	*need2free = 1;
	return buf;
#endif
}


/* C89 doesn't really allow converting between data and code pointers,
   but the platforms we target do. Centralize the casts here and explain
   gcc it is not a bug. */
#pragma GCC diagnostic ignored "-Wpedantic"

/* these casts are safe: they are protected by the checks below*/
void *stt_F2V(stt_fptr func)
{
	if (sizeof(void *) != sizeof(stt_fptr)) {
		fprintf(stderr, "estutter: data pointer and function pointer size differs\n");
		abort();
	}

	return (void *)func;
}

stt_fptr stt_V2F(void *v)
{
	if (sizeof(void *) != sizeof(stt_fptr)) {
		fprintf(stderr, "estutter: data pointer and function pointer size differs\n");
		abort();
	}

	return (stt_fptr)v;
}
